from ..interface import filer, searcher
from ..models.geometry import GeometricalTransformer
from ..models.interpreters import (
   INP_Interpreter,
   DAT_Interpreter,
   SVG_Interpreter
)

# Funções de Tradução
def inp_to_dat(input_data: str, flags: dict[str, str]):
   # Instanciando Interpretadores
   inp_interpreter = INP_Interpreter()
   dat_interpreter = DAT_Interpreter()

   # Interpretando Input
   inp_interpreter.read(input_data)

   # Transferindo Modelo de Simulação Interpretado
   dat_interpreter.model = inp_interpreter.model

   # Reordenando Nodes
   reference = searcher.get_database('translation_reference')['inp_to_dat']
   for group_ide, group in dat_interpreter.model.element_groups.items():
      # Idenfiticando Geometria do Grupo
      geometry = dat_interpreter.model.element_geometries[group.geometry_ide]
      
      # Pegando Reordenação da Referência
      nodes_reordering = reference['nodes_reordering'][geometry.shape][str(geometry.grade)]

      # Sobescrevendo Ordem dos Nodes
      for ide, element in group.elements.items():
         node_ides = [element.node_ides[i - 1] for i in nodes_reordering ]
         dat_interpreter.model.add_element(group_ide, ide, node_ides)

   # Retornando Tradução
   return dat_interpreter.write()

def dat_to_svg(input_data: str, flags: dict[str, str]):
   # Instanciando Interpretadores
   dat_interpreter = DAT_Interpreter()
   svg_interpreter = SVG_Interpreter()

   # Interpretando Input
   dat_interpreter.read_nodes(input_data)
   dat_interpreter.read_patches(input_data)
   dat_interpreter.read_elements(input_data)

   # Transferindo Modelos de Simulação
   svg_interpreter.model = dat_interpreter.model

   # Instanciando Transformador Geométrico e Configurações Padrões de Projeção
   gt = GeometricalTransformer()
   projection_type = 'parallel'
   rotations = list()

   # Identificando Configurações de Projeção a partir das Flags
   if len(flags) > 0:
      flag_keys = flags.keys()

      # Identificando Tipo de Projeção
      if ('-p' in flag_keys) or ('--projection' in flag_keys): 
         projection_type = flags.get('-p') or flags.get('--projection')
         if projection_type not in ('parallel', 'perspective'):
            raise ValueError(f'The projection type "{projection_type}" is not supported.')

         # Tratando Caso de Tipo de Projeção ser "perspective"
         if projection_type == 'perspective':
            try:
               x_cop = float(flags['-x'])
               y_cop = float(flags['-y'])
               z_cop = float(flags['-z'])
            except KeyError:
               raise ValueError(f'The projection type "perspective" requires 3 other flags: -x, -y and -z (the coordinates of the projection center).')

      # Lendo Rotações a Partir das Flags, se Ouverem
      for flag, value in flags.items():
         if (flag == '--Rx') or (flag == '--Ry') or (flag == '--Rz'):
            angle = float(value)
            rotations.append(
               gt.Rx(angle) if flag == '--Rx' else
               gt.Ry(angle) if flag == '--Ry' else
               gt.Rz(angle)
            )
   
   # Transportando Centroid para a Origem
   coordinates = [(n.x, n.y, n.z) for n in dat_interpreter.model.nodes.values()]
   centroid = gt.calculate_centroid(coordinates)
   coordinates = [gt.translate(*c, -centroid[0], -centroid[1], -centroid[2]) for c in coordinates]

   # Rotacionando Coordenadas
   coordinates = [gt.rotate(*c, rotations) for c in coordinates]

   # Projetando Coordenadas
   coordinates = [
      gt.project_parallel(*c)
      if projection_type == 'parallel' else
      gt.project_perspective(*c, x_cop, y_cop, z_cop)
      for c in coordinates
   ]
   u = [iso[0] for iso in coordinates]
   v = [iso[1] for iso in coordinates]

   # Ajustando Coordenadas ao Sistema SVG
   u_min = min(u)
   u_max = max(u)
   delta_u = u_max - u_min
   v_min = min(v)
   v_max = max(v)
   delta_v = v_max - v_min
   scale_coeff = (90 / delta_u) if delta_u > delta_v else (90 / delta_v)
   for (ide, node), u_i, v_i in zip(dat_interpreter.model.nodes.items(), u, v):
      # Ajustando Coordenadas para Eixo Padrão do SVG (Tudo Positivo | Eixo Vertical Invertido)
      x = u_i - u_min
      y = abs(v_i - v_max)

      # Ajustando Escala e Posição
      x = x * scale_coeff + 5
      y = y * scale_coeff + 5

      svg_interpreter.model.add_node(ide, x, y, 0, node.weight)

   # Retornando Tradução
   return svg_interpreter.write()

# Traduções Suportadas
supported_translations = {
   ('.inp', '.dat'): inp_to_dat,
   ('.dat', '.svg'): dat_to_svg
}

def start(input_path: str, output_extension: str, args: list[str], flags: dict[str, str]):
   # Lendo Arquivo de Input
   input_data = filer.read(input_path)

   # Verificando se a Tradução é Suportada
   last_dot_index = input_path.rfind('.')
   input_extension = input_path[last_dot_index:]
   format_pair = (input_extension, output_extension)
   try:
      translation_function = supported_translations[format_pair]
   except KeyError:
      raise KeyError(f'The translation of {input_extension} to {output_extension} is not supported.')
   
   # Traduzindo
   output_data = translation_function(input_data, flags)

   # Escrevendo Tradução no Output
   output_path = input_path[:last_dot_index] + output_extension
   filer.write(output_path, output_data)



import unittest
import os

class DefaultTest(unittest.TestCase):
   def default_test(self, benchmark_name: str, input_extension: str, output_extension: str, flags: str = ''):
      # Definindo paths
      benchmark_name = 'tests/benchmark/translate/' + benchmark_name
      input_path = benchmark_name + input_extension
      output_path = benchmark_name + output_extension
      exp_path = input_path[:-4] + '_exp' + output_extension

      # Traduzindo Benchmark
      code = os.system(f'python -m lmcv_tools translate {input_path} to {output_extension} {flags}')
      self.assertEqual(code, 0, 'A tradução falhou.')

      # Comparando Tradução com o Resultado Esperado
      output_file = open(output_path, 'r')
      exp_file = open(exp_path, 'r')
      output_data = output_file.read()
      exp_data = exp_file.read()
      output_file.close()
      exp_file.close()
      self.assertEqual(output_data, exp_data, 'A tradução está incorreta.')

      # Removendo Arquivo Gerado
      os.remove(output_path)

class Test_inp_to_dat(DefaultTest):
   def default_test(self, benchmark_name: str):
      super().default_test('inp_to_dat/' + benchmark_name, '.inp', '.dat')

   def test_Truss_T2D2(self):
      self.default_test('Truss_T2D2')

   def test_triangle_S3_1x1(self):
      self.default_test('Triangle_S3_1x1')

   def test_triangle_S3_2x2(self):
      self.default_test('Triangle_S3_2x2')

   def test_triangle_STRI65_1x1(self):
      self.default_test('Triangle_STRI65_1x1')

   def test_triangle_STRI65_2x2(self):
      self.default_test('Triangle_STRI65_2x2')

   def test_square_S4_1x1(self):
      self.default_test('Square_S4_1x1')
   
   def test_square_S4_2x2(self):
      self.default_test('Square_S4_2x2')

   def test_square_S8R_1x1(self):
      self.default_test('Square_S8R_1x1')
   
   def test_square_S8R_2x2(self):
      self.default_test('Square_S8R_2x2')
   
   def test_circle_S3_S4R_4x4(self):
      self.default_test('Circle_S3_S4R_4x4')

   def test_complex_part_S8R(self):
      self.default_test('ComplexPart_S8R')

   def test_cube_C3D4_12(self):
      self.default_test('Cube_C3D4_12')
   
   def test_cube_C3D10_12(self):
      self.default_test('Cube_C3D10_12')

   def test_cube_C3D8_1x1x1(self):
      self.default_test('Cube_C3D8_1x1x1')

   def test_cube_C3D8_2x2x2(self):
      self.default_test('Cube_C3D8_2x2x2')
   
   def test_cube_C3D20_1x1x1(self):
      self.default_test('Cube_C3D20_1x1x1')

   def test_cube_C3D20_2x2x2(self):
      self.default_test('Cube_C3D20_2x2x2')

   def test_complex_part_C3D20R(self):
      self.default_test('ComplexPart_C3D20R')
   
   def test_supported_square_S4(self):
      self.default_test('SupportedSquare_S4')

class Test_dat_to_svg(DefaultTest):
   def default_test(self, benchmark_name: str, flags: str = ''):
      super().default_test('dat_to_svg/' + benchmark_name, '.dat', '.svg', flags)

   def test_circle_T3_Q4_4x4(self):
      self.default_test('Circle_T3_Q4_4x4')

   def test_heart_plate_Q8_2x2(self):
      self.default_test('HeartPlate_Q8_2x2')

   def test_heart_plate_BT2_2x2(self):
      self.default_test('HeartPlate_BT2_2x2')

   def test_heart_plate_BT3_4x4(self):
      self.default_test('HeartPlate_BT3_4x4')

   def test_disform_BT2(self):
      self.default_test('Disform_BT2')
   
   def test_disform_BT4(self):
      self.default_test('Disform_BT4')
   
   def test_rectangle_NURBS_4x7(self):
      self.default_test('Rectangle_NURBS_4x7')
   
   def test_dirform_NURBS_2x2(self):
      self.default_test('Dirform_NURBS_2x2')

   def test_projection_plane_xy(self):
      self.default_test('CirclePlate_BT2_plane_xy', 'plane_xy')

   def test_projection_plane_yz(self):
      self.default_test('CirclePlate_BT2_plane_yz', '--Rz=-90.0 --Rx=-90.0')

   def test_projection_plane_xz(self):
      self.default_test('CirclePlate_BT2_plane_xz', '--Rx=-90.0')

   def test_projection_isometric(self):
      self.default_test('CirclePlate_BT2_isometric', '--Rz=45.0 --Rx=54.735610317')
   
   def test_projection_perspective(self):
      self.default_test('Hemisphere_Q8_perspective', '-p=perspective -x=0 -y=15 -z=40 --Ry=-20')


# --------------------------------------------------
# 1 - Matrix - Classes Relacionadas
# --------------------------------------------------

class Matrix:
    # Método Construtor
    def __init__(self, n_rows: int, n_columns: int):
        # Definindo Número de Linhas e Colunas
        self.n_rows = n_rows
        self.n_colmuns = n_columns

        # Inicializando Elementos
        self._elements = [[0 for j in range(n_columns)] for i in range(n_rows)]

    # Métodos Estáticos
    def from_list(l: list[list[float]]):
        M = Matrix(len(l), len(l[0]))
        M._elements = l
        return M

    # Métodos Públicos
    def fill(self, scalar: float):
        for i in range(self.n_rows):
            for j in range(self.n_colmuns):
                self[i, j] = scalar

    def fill_diagonal(self, scalar: float):
        for i in range(self.n_rows):
            for j in range(self.n_colmuns):
                if i == j:
                    self[i, j] = scalar

    def copy(self):
        copy_matrix = Matrix(self.n_rows, self.n_colmuns)
        copy_matrix._elements = [row.copy() for row in self._elements]
        return copy_matrix
    
    def to_list(self) -> list[list[int]]:
        return self._elements

    # Métodos Privados
    def _verify_index(self, index: tuple[int]):
        # Verificando Tamanho da Tupla
        if type(index) is not tuple or len(index) != 2:
            raise IndexError('The index tuple to get an item from an Matrix object must contain exactly 2 int indexes.')

    # Métodos Especiais
    def __str__(self) -> str:
        return f'Matrix(\n' + '\n'.join([
            '   ' + '   '.join([
                f'{self._elements[i][j]:.5e}'
                for j in range(self.n_colmuns)
            ])
            for i in range(self.n_rows)
        ]) + '\n)'

    def __getitem__(self, index: tuple[int]):
        self._verify_index(index)
        return self._elements[index[0]][index[1]]
    
    def __setitem__(self, index: tuple[int], scalar: float):
        self._verify_index(index)
        self._elements[index[0]][index[1]] = scalar

    def __add__(self, other):
        # Verificando se a Adição é apenas entre Matrizes
        if type(other) is not type(self):
            raise TypeError(f"Unsupported operand for +: {type(self).__name__}' and '{type(other).__name__}'")
        
        # Verificando se a Dimensão é a Mesma
        if self.n_rows != other.n_rows or self.n_colmuns != self.n_colmuns:
            raise ValueError(f'The Matrices must have the same dimensions.')
        
        # Criando Matriz de Soma
        sum_matrix = Matrix(self.n_rows, self.n_colmuns)
        for i in range(self.n_rows):
            for j in range(self.n_colmuns):
                sum_matrix[i, j] = self[i, j] + other[i, j]
        return sum_matrix
    
    def __mul__(self, other):
        # Tratando Caso do Outro Operando ser Int ou Float
        if type(other) is int or type(other) is float:
            mul_matrix = self.copy()
            for i in range(self.n_rows):
                for j in range(self.n_colmuns):
                    mul_matrix[i, j] *= other
            return mul_matrix
        
        # Tratando Caso do Outro Operando ser uma Matrix
        if type(other) is type(self):
            # Verificando de o N° de Colunas de O1 = N° de Colunas de O2 
            if self.n_colmuns != other.n_rows:
                raise ValueError('For Matrix Multiplication, the number of columns of the firts operand must be equal to number of rows of the second operand.')

            # Performando Multiplicação de Matrizes
            mul_matrix = Matrix(self.n_rows, other.n_colmuns)
            for i in range(self.n_rows):
                for j in range(other.n_colmuns):
                    for k in range(self.n_colmuns):
                        mul_matrix[i, j] += self[i, k] * other[k, j]
            return mul_matrix

        # Tratando Caso da Operação não ser Compatível
            raise TypeError(f"Unsupported operand for *: {type(self).__name__}' and '{type(other).__name__}'")
        
    def __rmul__(self, other):
        # Tratando Caso do Outro Operando ser Int ou Float
        if type(other) is int or type(other) is float:
            return self * other
        
    def __sub__(self, other):
        return other * (-1) + self



from ..interface import filer, messenger
from ..models.extraction_components import (
   ResultTable,
   Condition,
   Attribute
)

# Funções Globais
def extract(attributes: list[str], pos_data: str, condition: str = None) -> tuple[ResultTable, list]:
   # Instanciando Tabela de Resultado e Condição de Atributos
   result_table = ResultTable([])
   attributes_condition = Condition(condition)
   failed_atributes = list()
   
   # Analisando cada Atributo
   for attribute_name in attributes:
      # Instanciando Atributo
      attribute = Attribute(attribute_name)
      
      # Extraindo Dados Relativos ao Atributo dos Dados do .pos
      table = attribute.extract_from(pos_data, attributes_condition)

      # Cruzando Novos Dados com a Tabela de Resultado Final
      result_table = result_table.join(table)
      
      # Verificando se há Atributos não Cruzados
      if result_table.join_fail:
         failed_atributes.append(attribute_name)
   
   # Reordenando Dados (Não Falhos Apenas)
   valid_attributes = [a for a in attributes if a not in failed_atributes]
   result_table.reorder(valid_attributes)
   
   return result_table, failed_atributes

def start(attributes: list[str], pos_path: str, condition: str, csv_path: str):
   # Lendo Arquivo .pos
   pos_data = filer.read(pos_path)

   # Extraindo Dados e Convertendo para o Formato CSV
   table, failed_attributes = extract(attributes, pos_data, condition)
   csv_data = table.to_csv()

   # Informando Atributos não Cruzados
   if len(failed_attributes) > 0:
      messenger.warning('The following attributes could not be related to the others considering the attributes order informed:')
      messenger.show(','.join(failed_attributes))

   # Escrevendo Extração no .csv
   filer.write(csv_path, csv_data)


import unittest
import os

class DefaultTest(unittest.TestCase):
   def default_test(self, benchmark_name: str, benchmark_id: str, attribute: str, value: str, condition: str = None):
      # Definindo paths
      dat_path = 'tests/benchmark/change/' + benchmark_name + '.dat'
      temp_path = dat_path[:-4] + '_temp.dat'
      exp_path = dat_path[:-4] + '_exp_'+ benchmark_id + '.dat'

      # Copiando Dados para Arquivo Temporário de Teste
      dat_file = open(dat_path, 'r')
      dat_data = dat_file.read()
      dat_file.close()
      temp_file = open(temp_path, 'w')
      temp_file.write(dat_data)
      temp_file.close()
      del dat_data

      # Alterando Atributo
      command = f'python -m lmcv_tools change {attribute} to {value} in {temp_path}'
      if condition:
         command += f' where {condition}'
      code = os.system(command)
      self.assertEqual(code, 0, 'A alteração falhou.')

      # Comparando Alteração com o Resultado Esperado
      temp_file = open(temp_path, 'r')
      exp_file = open(exp_path, 'r')
      temp_data = temp_file.read()
      exp_data = exp_file.read()
      temp_file.close()
      exp_file.close()
      self.assertEqual(temp_data, exp_data, 'A alteração está incorreta.')

      # Removendo Arquivo Temporário Gerado
      os.remove(temp_path)

class TestUnconditionedChanges(DefaultTest):
   def test_material_isotropic_E(self):
      self.default_test(
         'ThreeBarTruss', 
         'material_isotropic_E',
         'material.isotropic.E',
         '300.0e+09'
      )
   
   def test_material_isotropic_poisson(self):
      self.default_test(
         'ThreeBarTruss', 
         'material_isotropic_poisson',
         'material.isotropic.poisson',
         '0.5'
      )

   def test_section_bar_circle_radius(self):
      self.default_test(
         'ThreeBarTruss', 
         'section_bar_circle_radius',
         'section.bar.circle.radius',
         '2.40000e-03'
      )

class TestConditionedChanges(DefaultTest):
   def test_condition_1(self):
      self.default_test(
         'ThreeBarTruss', 
         'condition_1',
         'section.bar.circle.radius',
         '2.40000e-03',
         'section.id = 2'
      )


import unittest
import os

class DefaultTest(unittest.TestCase):
   def default_test(self, method: str, benchmark_name: str, benchmark_id: str = None):
      # Definindo paths
      dat_path = 'tests/benchmark/reorder/' + benchmark_name + '.dat'
      temp_path = dat_path[:-4] + '_temp.dat'
      exp_path = dat_path[:-4] + '_exp'
      if benchmark_id:
         exp_path += '_'+ benchmark_id
      exp_path += '.dat'

      # Copiando Dados para Arquivo Temporário de Teste
      dat_file = open(dat_path, 'r')
      dat_data = dat_file.read()
      dat_file.close()
      temp_file = open(temp_path, 'w')
      temp_file.write(dat_data)
      temp_file.close()
      del dat_data

      # Executando Reordenação
      command = f'python -m lmcv_tools reorder {method} {temp_path}'
      code = os.system(command)
      self.assertEqual(code, 0, 'A reordenação falhou.')

      # Comparando Reordenação com o Resultado Esperado
      exp_file = open(exp_path, 'r')
      exp_data = exp_file.read()
      exp_file.close()
      temp_file = open(temp_path, 'r')
      temp_data = temp_file.read()
      temp_file.close()
      self.assertEqual(temp_data, exp_data, 'A reordenação está incorreta.')

      # Removendo Arquivo Temporário Gerado
      os.remove(temp_path)

class TestMethods(DefaultTest):
   def test_rcm(self):
      benchmark = ('ComplexPipe', 'rcm')
      self.default_test('rcm', *benchmark)
   
   def test_sloan(self):
      benchmark = ('ComplexPipe', 'sloan')
      self.default_test('sloan', *benchmark)

   def test_boost_rcm(self):
      benchmark = ('ComplexPipe', 'boost_rcm')
      self.default_test('boost_rcm', *benchmark)

   def test_boost_sloan(self):
      benchmark = ('ComplexPipe', 'boost_sloan')
      self.default_test('boost_sloan', *benchmark)

class TestBases(DefaultTest):
   def test_lagrange(self):
      self.default_test('rcm', 'SquareLagrange')
   
   def test_bezier_triangle(self):
      self.default_test('rcm', 'SquareBezierTriangle')

   def test_bezier_surface(self):
      self.default_test('rcm', 'SquareBezierSurface')
   
   def test_bspline_surface(self):
      self.default_test('rcm', 'SquareBsplineSurface')
   


import unittest
import os

class DefaultTest(unittest.TestCase):
   def default_test(self, benchmark_name: str, benchmark_id: str, attributes: list[str], condition: str):
      # Definindo paths
      pos_path = 'tests/benchmark/extract/' + benchmark_name + '.pos'
      csv_path = pos_path[:-3] + 'csv'
      exp_path = pos_path[:-4] + '_exp_'+ benchmark_id + '.csv'

      # Extraindo Dados
      attributes_joined = ' '.join(attributes)
      command = f'python -m lmcv_tools extract {attributes_joined} from {pos_path}'
      if condition:
         command += f' where {condition}'
      code = os.system(command)
      self.assertEqual(code, 0, 'A extração falhou.')

      # Comparando Extração com o Resultado Esperado
      csv_file = open(csv_path, 'r')
      exp_file = open(exp_path, 'r')
      csv_data = csv_file.read()
      exp_data = exp_file.read()
      csv_file.close()
      exp_file.close()
      self.assertEqual(csv_data, exp_data, 'A extração está incorreta.')

      # Removendo Arquivo .csv Gerado
      os.remove(csv_path)

class TestSingleAttributes(DefaultTest):
   def test_step_id(self):
      benchmark = ('CirclePlate_Plastic', 'step_id')
      attributes = ['step.id']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_factor(self):
      benchmark = ('CirclePlate_Plastic', 'step_factor')
      attributes = ['step.factor']
      condition = ''
      self.default_test(*benchmark, attributes, condition)

   def test_node_id(self):
      benchmark = ('CirclePlate_Plastic', 'node_id')
      attributes = ['node.id']
      condition = ''
      self.default_test(*benchmark, attributes, condition)

   def test_node_coord_0(self):
      benchmark = ('CirclePlate_Plastic', 'node_coord_0')
      attributes = ['node.coord[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_node_coord_1(self):
      benchmark = ('CirclePlate_Plastic', 'node_coord_1')
      attributes = ['node.coord[1]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_node_coord_2(self):
      benchmark = ('CirclePlate_Plastic', 'node_coord_2')
      attributes = ['node.coord[2]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_node_displacement_0(self):
      benchmark = ('CirclePlate_Plastic', 'step_node_displacement_0')
      attributes = ['step.node.displacement[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_node_displacement_1(self):
      benchmark = ('CirclePlate_Plastic', 'step_node_displacement_1')
      attributes = ['step.node.displacement[1]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_node_displacement_2(self):
      benchmark = ('CirclePlate_Plastic', 'step_node_displacement_2')
      attributes = ['step.node.displacement[2]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_element_id(self):
      benchmark = ('CirclePlate_Plastic', 'element_id')
      attributes = ['element.id']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_element_info_0(self):
      benchmark = ('CirclePlate_Plastic', 'element_info_0')
      attributes = ['element.info[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_element_info_2(self):
      benchmark = ('CirclePlate_Plastic', 'element_info_2')
      attributes = ['element.info[2]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_gauss_scalar_0(self):
      benchmark = ('CirclePlate_Plastic', 'step_gauss_scalar_0')
      attributes = ['step.gauss.scalar[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_gauss_scalar_3(self):
      benchmark = ('CirclePlate_Plastic', 'step_gauss_scalar_3')
      attributes = ['step.gauss.scalar[3]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)

   def test_step_element_gauss_scalar_data_0(self):
      benchmark = ('CirclePlate_Plastic', 'step_element_gauss_scalar_data_0')
      attributes = ['step.element.gauss.scalar.data[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_element_gauss_scalar_data_15(self):
      benchmark = ('CirclePlate_Plastic', 'step_element_gauss_scalar_data_15')
      attributes = ['step.element.gauss.scalar.data[15]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)

   def test_step_nodal_scalar_0(self):
      benchmark = ('CirclePlate_Plastic', 'step_nodal_scalar_0')
      attributes = ['step.nodal.scalar[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_nodal_scalar_3(self):
      benchmark = ('CirclePlate_Plastic', 'step_nodal_scalar_3')
      attributes = ['step.nodal.scalar[3]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)

   def test_step_element_nodal_scalar_data_0(self):
      benchmark = ('CirclePlate_Plastic', 'step_element_nodal_scalar_data_0')
      attributes = ['step.element.nodal.scalar.data[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_step_element_nodal_scalar_data_31(self):
      benchmark = ('CirclePlate_Plastic', 'step_element_nodal_scalar_data_31')
      attributes = ['step.element.nodal.scalar.data[31]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_buckling_id(self):
      benchmark = ('SquarePlate', 'buckling_id')
      attributes = ['buckling.id']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_buckling_factor(self):
      benchmark = ('SquarePlate', 'buckling_factor')
      attributes = ['buckling.factor']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_buckling_node_displacement_0(self):
      benchmark = ('SquarePlate', 'buckling_node_displacement_0')
      attributes = ['buckling.node.displacement[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_buckling_node_displacement_1(self):
      benchmark = ('SquarePlate', 'buckling_node_displacement_1')
      attributes = ['buckling.node.displacement[1]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_buckling_node_displacement_2(self):
      benchmark = ('SquarePlate', 'buckling_node_displacement_2')
      attributes = ['buckling.node.displacement[2]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_vibration_id(self):
      benchmark = ('HeartPlate', 'vibration_id')
      attributes = ['vibration.id']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_vibration_factor(self):
      benchmark = ('HeartPlate', 'vibration_factor')
      attributes = ['vibration.factor']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_vibration_node_displacement_0(self):
      benchmark = ('HeartPlate', 'vibration_node_displacement_0')
      attributes = ['vibration.node.displacement[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_vibration_node_displacement_1(self):
      benchmark = ('HeartPlate', 'vibration_node_displacement_1')
      attributes = ['vibration.node.displacement[1]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_vibration_node_displacement_2(self):
      benchmark = ('HeartPlate', 'vibration_node_displacement_2')
      attributes = ['vibration.node.displacement[2]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)

class TestMultipleAttributes(DefaultTest):
   def test_multiple_1(self):
      benchmark = ('CirclePlate_Plastic', 'multiple_1')
      attributes = ['step.id', 'step.factor']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_multiple_2(self):
      benchmark = ('CirclePlate_Plastic', 'multiple_2')
      attributes = ['step.factor', 'step.node.displacement[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_multiple_3(self):
      benchmark = ('CirclePlate_Plastic', 'multiple_3')
      attributes = ['step.node.displacement[0]', 'node.id', 'node.coord[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_multiple_unrelated_1(self):
      benchmark = ('CirclePlate_Plastic', 'multiple_ur_1')
      attributes = ['node.id', 'step.id']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
   def test_multiple_unrelated_2(self):
      benchmark = ('CirclePlate_Plastic', 'multiple_ur_2')
      attributes = ['node.id', 'step.id', 'node.coord[0]']
      condition = ''
      self.default_test(*benchmark, attributes, condition)
   
class TestCondition(DefaultTest):
   def test_condition_equal(self):
      benchmark = ('CirclePlate_Plastic', 'cond_eq')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id = 5'
      self.default_test(*benchmark, attributes, condition)
   
   def test_condition_less(self):
      benchmark = ('CirclePlate_Plastic', 'cond_ls')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id "<" 5'
      self.default_test(*benchmark, attributes, condition)
   
   def test_condition_greater(self):
      benchmark = ('CirclePlate_Plastic', 'cond_gt')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id ">" 5'
      self.default_test(*benchmark, attributes, condition)
   
   def test_condition_less_or_equal(self):
      benchmark = ('CirclePlate_Plastic', 'cond_le')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id "<=" 5'
      self.default_test(*benchmark, attributes, condition)
   
   def test_condition_greater_or_equal(self):
      benchmark = ('CirclePlate_Plastic', 'cond_ge')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id ">=" 5'
      self.default_test(*benchmark, attributes, condition)
   
   def test_condition_different(self):
      benchmark = ('CirclePlate_Plastic', 'cond_df')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id "!=" 5'
      self.default_test(*benchmark, attributes, condition)
   
   def test_condition_in_1(self):
      benchmark = ('CirclePlate_Plastic', 'cond_in_1')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id in 1:4'
      self.default_test(*benchmark, attributes, condition)
   
   def test_condition_in_2(self):
      benchmark = ('CirclePlate_Plastic', 'cond_in_2')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id in 1:11:2'
      self.default_test(*benchmark, attributes, condition)

   def test_condition_and(self):
      benchmark = ('CirclePlate_Plastic', 'cond_and')
      attributes = ['step.id', 'step.factor']
      condition = 'step.id "<" 9 and step.factor ">" 1.6'
      self.default_test(*benchmark, attributes, condition)
   
   def test_condition_or(self):
      benchmark = ('CirclePlate_Plastic', 'cond_or')
      attributes = ['step.id', 'step.factor']
      condition = 'step.factor "<" 0.5 or step.factor ">" 2.5'
      self.default_test(*benchmark, attributes, condition)


from tkinter import Tk, filedialog, END
from tkinter.ttk import (
   Style,
   Frame,
   Label,
   Entry,
   Spinbox,
   Combobox,
   Button,
   Checkbutton
)
from ..interface import messenger

class PromptGenerateVirtualLaminas:
   def __init__(self, params: dict):
      # Dicionário de Parâmetros para Geração das Lâminas
      self.params = params

      # Definindo Janela Principal
      self.root = Tk()
      self.root.title('Generate Virtual Laminas')
      self.root.resizable(False, False)
      self.root.geometry('+200+100')

      # Definindo Organização de Widgets
      self.frame = dict()
      self.label = dict()
      self.entry = dict()
      self.button = dict()

      # Estilos
      self.default_font = ('sans-serif', 12)
      self.style = Style(self.root)
      self.style.configure('TLabel', font = self.default_font)
      self.style.configure('TButton', font = self.default_font)
      self.style.configure('title.TLabel', font = ('sans-serif', 14, 'italic'))

      # Definindo Layout
      self.layout_laminar_params()
      self.layout_grading_params()
      self.layout_materials_params()
      self.layout_action()
   
   def layout_laminar_params(self):
      self.frame['laminar_params'] = Frame(
         master = self.root,
         padding = (10, 10),
      )
      self.frame['laminar_params'].pack(fill = 'x')

      self.label['title_laminar'] = Label(
         master = self.frame['laminar_params'],
         text = 'Laminar Parameters',
         style = 'title.TLabel',
         padding = (0, 0, 0, 10),
      )
      self.label['title_laminar'].grid(row = 0, columnspan = 4, sticky = 'ws')

      self.label['laminas_count'] = Label(
         master = self.frame['laminar_params'], 
         text = 'Laminas Count:',
         padding = (0, 5, 10, 5),
      )
      self.label['laminas_count'].grid(row = 1, column = 0, sticky = 'w')

      self.entry['laminas_count'] = Spinbox(
         master = self.frame['laminar_params'], 
         from_ = 1,
         to = 100.0,
         width = 3,
         font = self.default_font,
      )
      self.entry['laminas_count'].grid(row = 1, column = 1, sticky = 'w')

      self.label['element_type'] = Label(
         master = self.frame['laminar_params'],
         text = 'Element Type: ',
         padding = (0, 5, 0, 5),
      )
      self.label['element_type'].grid(row = 2, column = 0, sticky = 'w')

      self.entry['element_type'] = Combobox(
         master = self.frame['laminar_params'],
         state = 'readonly',
         values = ['Solid', 'Shell'],
         font = self.default_font,
         width = 12,
      )
      self.entry['element_type'].bind('<<ComboboxSelected>>', self.change_element_type)
      self.entry['element_type'].grid(row = 2, column = 1, sticky = 'w')

      self.label['total_thickness'] = Label(
         master = self.frame['laminar_params'], 
         text = 'Laminas Total\nThickness:',
         padding = (0, 5, 10, 5)
      )

      self.entry['total_thickness'] = Spinbox(
         master = self.frame['laminar_params'], 
         from_ = 0.1,
         to = 100,
         increment = 0.1,
         width = 6,
         font = self.default_font,
      )

      self.label['total_elements'] = Label(
         master = self.frame['laminar_params'], 
         text = 'Total Elements\nin Thickness:',
         padding = (0, 5, 10, 5)
      )

      self.entry['total_elements'] = Spinbox(
         master = self.frame['laminar_params'], 
         from_ = 1,
         to = 100,
         increment = 1,
         width = 6,
         font = self.default_font,
      )

      self.label['number_integration_points'] = Label(
         master = self.frame['laminar_params'], 
         text = 'Number of Integration\nPoints per Lamina:',
         padding = (0, 5, 10, 5)
      )
      self.label['number_integration_points'].grid(row = 4, column = 0, sticky = 'w')

      self.entry['number_integration_points'] = Spinbox(
         master = self.frame['laminar_params'], 
         from_ = 1,
         to = 15,
         increment = 1,
         width = 6,
         font = self.default_font,
      )
      self.entry['number_integration_points'].grid(row = 4, column = 1, sticky = 'w')

      self.entry['smart'] = Checkbutton(
         master = self.frame['laminar_params'],
         text = "Smart Laminas",
         command = self.check_smart_laminas
      )
      self.entry['smart'].grid(row = 5, column = 0, sticky = 'w')

   def layout_grading_params(self):
      self.frame['grading_params'] = Frame(
         master = self.root,
         padding = (10, 10),
      )
      self.frame['grading_params'].pack(fill = 'x')

      self.label['title_grading'] = Label(
         master = self.frame['grading_params'],
         text = 'Functionally Grading Parameters',
         style = 'title.TLabel',
         padding = (0, 0, 0, 10),
      )
      self.label['title_grading'].grid(row = 0, columnspan = 2, sticky = 'w')

      self.label['power_law_exponent'] = Label(
         master = self.frame['grading_params'], 
         text = 'Power Law Exponent:',
         padding = (0, 5, 10, 5)
      )
      self.label['power_law_exponent'].grid(row = 1, column = 0, sticky = 'w')

      self.entry['power_law_exponent'] = Spinbox(
         master = self.frame['grading_params'], 
         from_ = 0,
         to = 10,
         increment = 0.1,
         width = 6,
         font = self.default_font,
      )
      self.entry['power_law_exponent'].grid(row = 2, column = 0, sticky = 'w')

      self.label['micromechanical_model'] = Label(
         master = self.frame['grading_params'], 
         text = 'Micromechanical Model:',
         padding = (0, 5, 10, 5),
      )
      self.label['micromechanical_model'].grid(row = 3, column = 0, sticky = 'w')

      self.entry['micromechanical_model'] = Combobox(
         master = self.frame['grading_params'],
         state = 'readonly',
         values = ['voigt', 'mori_tanaka', 'hashin_shtrikman_lower_bound', 'hashin_shtrikman_upper_bound'],
         font = self.default_font,
         width = 25,
      )
      self.entry['micromechanical_model'].grid(row = 4, column = 0, sticky = 'w')

   def layout_materials_params(self):
      self.frame['materials_params'] = Frame(
         master = self.root,
         padding = (10, 10),
      )
      self.frame['materials_params'].pack(fill = 'x')

      self.label['title_materials'] = Label(
         master = self.frame['materials_params'],
         text = 'Materials Parameters',
         style = 'title.TLabel',
         padding = (0, 0, 0, 10),
      )
      self.label['title_materials'].grid(row = 0, columnspan = 6, sticky = 'ws')

      self.label['E1'] = Label(
         master = self.frame['materials_params'],
         text = 'E1 = ',
         padding = (0, 5, 0, 5),
      )
      self.label['E1'].grid(row = 1, column = 0, sticky = 'w')

      self.entry['E1'] = Entry(
         master = self.frame['materials_params'], 
         width = 6,
         font = self.default_font,
      )
      self.entry['E1'].grid(row = 1, column = 1, sticky = 'w', padx = 5)

      self.label['nu1'] = Label(
         master = self.frame['materials_params'],
         text = 'ν1 = ',
         padding = (0, 5, 0, 5),
      )
      self.label['nu1'].grid(row = 1, column = 2, sticky = 'w', padx = 5)

      self.entry['nu1'] = Entry(
         master = self.frame['materials_params'], 
         width = 6,
         font = self.default_font,
      )
      self.entry['nu1'].grid(row = 1, column = 3, sticky = 'w', padx = 5)

      self.label['rho1'] = Label(
         master = self.frame['materials_params'],
         text = 'ρ1 = ',
         padding = (0, 5, 0, 5),
      )
      self.label['rho1'].grid(row = 1, column = 4, sticky = 'w', padx = 5)

      self.entry['rho1'] = Entry(
         master = self.frame['materials_params'], 
         width = 6,
         font = self.default_font,
      )
      self.entry['rho1'].grid(row = 1, column = 5, sticky = 'w')

      self.label['E2'] = Label(
         master = self.frame['materials_params'],
         text = 'E2 = ',
         font = self.default_font,
         padding = (0, 5, 0, 5),
      )
      self.label['E2'].grid(row = 2, column = 0, sticky = 'w')

      self.entry['E2'] = Entry(
         master = self.frame['materials_params'], 
         width = 6,
         font = self.default_font,
      )
      self.entry['E2'].grid(row = 2, column = 1, sticky = 'w', padx = 5)

      self.label['nu2'] = Label(
         master = self.frame['materials_params'],
         text = 'ν2 = ',
         padding = (0, 5, 0, 5),
      )
      self.label['nu2'].grid(row = 2, column = 2, sticky = 'w', padx = 5)

      self.entry['nu2'] = Entry(
         master = self.frame['materials_params'], 
         width = 6,
         font = self.default_font,
      )
      self.entry['nu2'].grid(row = 2, column = 3, sticky = 'w', padx = 5)

      self.label['rho2'] = Label(
         master = self.frame['materials_params'],
         text = 'ρ2 = ',
         padding = (0, 5, 0, 5),
      )
      self.label['rho2'].grid(row = 2, column = 4, sticky = 'w', padx = 5)

      self.entry['rho2'] = Entry(
         master = self.frame['materials_params'], 
         width = 6,
         font = self.default_font,
      )
      self.entry['rho2'].grid(row = 2, column = 5, sticky = 'w')

   def layout_action(self):
      self.frame['action'] = Frame(
         master = self.root,
         padding = (10, 10),
      )
      self.frame['action'].pack(fill = 'x')

      self.label['path'] = Label(
         master = self.frame['action'],
         text = 'Output File Path:',
         padding = (0, 5, 0, 5),
      )
      self.label['path'].grid(row = 0, column = 0, sticky = 'w')

      self.entry['path'] = Entry(
         master = self.frame['action'], 
         font = self.default_font,
         width = 25,
      )
      self.entry['path'].grid(row = 1, column = 0, )

      self.button['path'] = Button(
         master = self.frame['action'],
         text = 'Select Folder',
         command = self.select_path,
      )
      self.button['path'].grid(row = 1, column = 1)

      self.button['generate'] = Button(
         master = self.frame['action'],
         text = 'Generate',
         command = self.validate,
      )
      self.button['generate'].grid(row = 2, column = 0, columnspan = 2, pady=(5, 0))

   def start(self):
      self.root.mainloop()

   def select_path(self):
      path = filedialog.askdirectory()
      if path:
         self.entry['path'].delete(0, END)
         self.entry['path'].insert(0, path)

   def validate(self):
      # Copiando Entradas para Validação
      entries = self.entry.copy()

      # Verificando Tipo de Elemento
      if (element_type := entries['element_type'].get()) == '':
         return messenger.show('"Element Type" is required.')
      self.params['element_type'] = element_type

      # Verificando Espessura com Base no Tipo do Elemento
      if element_type == 'Shell':
         if (thickness := entries['total_thickness'].get()) == '':
            return messenger.show('"Laminas Absolute Thickness" is required.')
         self.params['thickness'] = float(thickness)
      else:
         if (total_elements := entries['total_elements'].get()) == '':
            return messenger.show('"Laminas Per Element" is required.')
         self.params['total_elements'] = float(total_elements)

      # Verificando Modelo Micromecânico
      if (micromechanical_model := entries['micromechanical_model'].get()) == '':
         return messenger.show('"Micromechanical Model" is required.')
      self.params['micromechanical_model'] = micromechanical_model

      # Verificando se a Distribuição deve ser Adaptativa
      self.params['smart'] = 'selected' in entries['smart'].state()

      # Atribuindo Caminho do Arquivo
      self.params['path'] = entries['path'].get().strip()

      # Deletando Entradas Verificadas
      del entries['element_type']
      del entries['total_elements']
      del entries['total_thickness']
      del entries['micromechanical_model']
      del entries['smart']
      del entries['path']

      # Verificando Demais Entradas
      for key, value in entries.items():
         if value.get() == '':
            return messenger.show('All fieds are required (except "Output File Path"), fill them.')
         type_class = int if key in ['laminas_count', 'number_integration_points'] else float
         self.params[key] = type_class(value.get())
      
      # Gerando Espessura Relativa para Elementos Sólidos
      if self.params.get('total_elements'):
         self.params['thickness'] = self.params['total_elements'] / self.params['laminas_count']
         del self.params['total_elements']
      
      # Terminando Interface
      self.root.destroy()
   
   def check_smart_laminas(self):
      state = self.entry['smart'].state()
      if 'selected' in state:
         self.entry['total_elements'].set(1)
         self.entry['total_elements'].config(state = 'disabled')
      else:
         self.entry['total_elements'].config(state = 'enabled')

   def change_element_type(self, e):
      value = self.entry['element_type'].get()
      if value == 'Shell':
         self.label['total_elements'].grid_forget()
         self.entry['total_elements'].grid_forget()
         self.label['total_thickness'].grid(row = 3, column = 0, sticky = 'w')
         self.entry['total_thickness'].grid(row = 3, column = 1, sticky = 'w')
      else:
         self.label['total_thickness'].grid_forget()
         self.entry['total_thickness'].grid_forget()
         self.label['total_elements'].grid(row = 3, column = 0, sticky = 'w')
         self.entry['total_elements'].grid(row = 3, column = 1, sticky = 'w')



from math import sin, cos, pi
from ..interface import searcher
from .simulation import (
   SimulationModel,
   FunctionallyGradedMaterial
)
from .interpreters import (
   DAT_Interpreter
)
from .geometry import NURBS_Curve

# --------------------------------------------------
# 1 - Classe Abstrata de Artefato
# --------------------------------------------------
class Artifact:
   def __init__(self, name: str, file_extension: str, data: str = ''):
      self.name = name
      self.file_extension = file_extension
      self.data = data
   
   @property
   def file_name(self) -> str:
      return self.name + '.' + self.file_extension
   
   # Para Implementar
   def generate(self):
      return self.data

# --------------------------------------------------
# 2 - Classes do Artefato "virtual_laminas"
# --------------------------------------------------
class ElementConfiguration:
   # Elementos Suportados
   supported_types = {'Solid', 'Shell'}

   def __init__(self, type: str, number_integration_points: int):
      if type not in ElementConfiguration.supported_types:
         raise ValueError(f'Element Type "{type}" is not supported.')
      self.type = type
      self.number_integration_points = number_integration_points

class VirtualLaminas(Artifact):
   def __init__(
      self,
      laminas_count: int,
      thickness: float,
      power_law_exponent: float,
      element_configuration: ElementConfiguration,
      fgm: FunctionallyGradedMaterial,
      smart: bool = False
   ):
      super().__init__('virtual_laminas', 'inp')
      self.laminas_count = laminas_count
      self.thickness = thickness
      self.power_law_exponent = power_law_exponent
      self.element_configuration = element_configuration
      self.fgm = fgm
      self.smart = smart
   
   def volume_fraction(self, z: float):
      return 1 - z ** self.power_law_exponent
   
   def z_coordinate(self, V: float):
      return (1 - V) ** (1 / self.power_law_exponent)

   def same_thickness_laminas(self):
      step = 1 / self.laminas_count
      points = [step / 2 + i * step for i in range(self.laminas_count)]
      fractions = [self.volume_fraction(z) for z in points]
      if self.element_configuration.type == 'Solid':
         thickness = [self.thickness for _ in range(self.laminas_count)]
      else:
         thickness = [step * self.thickness for _ in range(self.laminas_count)]
      return fractions, thickness

   def smart_laminas(self):
      # Variáveis Iniciais
      n = self.laminas_count
      p = self.power_law_exponent
      V = self.volume_fraction
      z = self.z_coordinate
      fractions_z = list()
      fractions_V = list()
      thickness_z = list()
      thickness_V = list()
      
      # Calculando Z de Referência
      if p == 1:
         z_ref = 0.5
      else:
         z_ref = p ** (-1 / (p - 1))
      
      # Decidindo se a Região de Prioridade z está à Esquerda ou Direita
      V_ref = V(z_ref)
      slope_tendency = -p * (p - 1) * z_ref ** (p - 2)
      if slope_tendency > 0:
         l_V = 1 - V_ref
         l_z = 1 - z_ref
      else:
         l_V = V_ref
         l_z = z_ref

      # Parâmetros da Região de Prioridade V
      n_V = round(l_z * n)
      step_V = l_V / n_V
      if slope_tendency > 0:
         z_0 = 0
         V_i = 1 - step_V / 2
      else:
         z_0 = z_ref
         V_i = V_ref - step_V / 2

      # Gerando Laminas da Região de Prioridade V 
      for _ in range(n_V):
         # Calculando Espessura Variável
         h_i = z(V_i - step_V / 2) - z_0

         # Registrando Informações
         fractions_V.append(V_i)
         thickness_V.append(h_i)

         # Atualizando Fração de Volume e Referência para Espessura
         V_i -= step_V
         z_0 += h_i
      
      # Parâmetros da Região de Prioridade z
      n_z = n - n_V
      step_z = l_z / n_z
      if slope_tendency > 0:
         z_i = z_ref + step_z / 2
      else:
         z_i = step_z / 2

      # Gerando Laminas da Região de Prioridade z
      for _ in range(n_z):
         fractions_z.append(V(z_i))
         thickness_z.append(step_z)
         z_i += step_z
      
      # Mesclando Regiões
      if slope_tendency > 0:
         fractions = fractions_V + fractions_z
         thickness = thickness_V + thickness_z
      else:
         fractions = fractions_z + fractions_V
         thickness = thickness_z + thickness_V

      # Corrigindo Espessura
      thickness = [t * self.thickness for t in thickness]

      return fractions, thickness

   def generate(self):
      # Inicializando Dados
      inp_data = ''

      # Gerados Dados de Lâminas
      laminas = self.smart_laminas() if self.smart else self.same_thickness_laminas()

      # Escrevendo Materiais
      material_names = list()
      index = 1
      for V in laminas[0]:
         # Gerando e Armazando Nome de Material
         name =  f'FGM-L{index}'
         material_names.append(name)

         # Homogeneizando Propriedades
         E, nu, rho = self.fgm.homogenize([V, 1 - V])

         # Adicionando Dados
         inp_data += f'*Material, name={name}\n    *Density\n    {rho:.7E},\n    *Elastic\n    {E:.7E}, {nu:.3f}\n'
         
         index += 1
      
      # Preparando para Escrever Lâminas
      inp_data += '*Part, name=Virtual_Part\n*Node\n    1, 1.0, 1.0, 0.0\n    2, 0.0, 1.0, 0.0\n    3, 0.0, 0.0, 0.0\n    4, 1.0, 0.0, 0.0\n*Element, type=S4R\n    1, 1, 2, 3, 4\n*Elset, elset=Virtual\n    1'
      element_type = self.element_configuration.type
      int_points = self.element_configuration.number_integration_points
      rotation_angle = 0

      # Escrevendo Lâmina por Lâmina
      inp_data += f'\n*{element_type} Section, elset=Virtual, composite\n'
      index = 1
      for h, material in zip(laminas[1], material_names):
         inp_data += f'    {h:.7E}, {int_points}, {material}, {rotation_angle}, Ply-{index}\n'
         index += 1
      inp_data += '*End Part'

      # Inseridos dados Inp no Atributo de Dados
      self.data = inp_data

# --------------------------------------------------
# 3 - Classes do Artefato "rectangle"
# --------------------------------------------------
class Rectangle(Artifact):
   # Funções de Geração de Coordenadas e Incidência de Elementos
   def _q8_coordinates(self):
      # Renomeando Atributos
      width, height = self.dimensions
      nx, ny = self.discretization

      # Calculando Valores Necessários
      delta_x = width / nx
      delta_y = height / ny
      x_values = [delta_x / 2 * i for i in range(2 * nx + 1)]
      y_values = [delta_y / 2 * i for i in range(2 * ny + 1)]

      # Gerando Coordenadas
      ide = 1
      for i_y, y in enumerate(y_values):
         # Mudando Valores de x com base em y
         xs = x_values if i_y % 2 == 0 else x_values[::2]

         for x in xs:
            self.model.add_node(ide, x, y, 0.0)
            ide += 1

   def _q8_incidence(self, i: int) -> list[int]:
      # Inicializando Variáveis
      nx, _ = self.discretization
      inc = [0] * 8
      x_order = i % nx
      if x_order == 0: x_order = nx
      y_order = i // nx
      if i % nx != 0: y_order += 1
      
      # Determinando Nó Inicial
      inc[0] = (2 * x_order - 1) + (y_order - 1) * (3 * nx + 2)

      # Determinando Restante da Incidência
      inc[1] = inc[0] + 1
      inc[2] = inc[0] + 2
      inc[3] = inc[0] + 2 * nx + 3 - x_order
      inc[4] = inc[3] + nx + 1 + x_order
      inc[5] = inc[4] - 1
      inc[6] = inc[4] - 2
      inc[7] = inc[3] - 1

      # Ordenando Conforme o FAST
      fast_order = [0, 1, 2, 3, 4, 5, 6, 7]
      inc = [inc[i] for i in fast_order]

      return inc

   # Elementos Suportados
   supported_elements = {
      'Q8': {
         'coordinates': _q8_coordinates,
         'incidence': _q8_incidence
      }
   }

   def __init__(
      self,
      element_type: str,
      dimensions: list[float],
      discretization: list[int]
   ):
      # Chamando Construtor da Superclasse
      super().__init__('rectangle', 'dat')

      # Verificando se Tipo de Elemento Fornecido é Suportado
      if element_type not in Rectangle.supported_elements.keys():
         raise ValueError(f'Element Type "{element_type}" is not supported for rectangle generation.')   

      # Verificando se o Número de Dimensões e Discretização foram passadas Corretamente
      if len(dimensions) != 2:
         raise ValueError('A Rectangle needs exactly 2 dimensions (width and height).')
      if len(discretization) != 2:
         raise ValueError('A Rectangle needs exactly 2 discretization values (number of elements in width and height).')

      # Atribuindo Atributos
      self.element_type = element_type
      self.dimensions = dimensions
      self.discretization = discretization
      self.model = SimulationModel()
      self._coordinates = Rectangle.supported_elements[element_type]['coordinates']
      self._incidence = Rectangle.supported_elements[element_type]['incidence']
      self.reference = searcher.get_database('translation_reference')

   def coordinates(self):
      return self._coordinates(self)
   
   def incidence(self, i: int) -> list[int]:
      return self._incidence(self, i)
   
   def geometry(self) -> int:
      element_info = self.reference['dat']['elements'][self.element_type]
      return self.model.add_element_geometry(**element_info)

   def generate(self):
      # Renomeando Atributos
      nx, ny = self.discretization

      # Gerando Coordenadas e Nodes
      self.coordinates()

      # Configurações dos Elementos
      geometry_ide = self.geometry()
      self.model.add_element_group(1, geometry_ide, None)

      # Gerando Elementos
      for i in range(nx * ny):
         # Gerando Incidência
         nodal_incidence = self.incidence(i + 1)

         # Cadastrando Elemento
         self.model.add_element(
            group_ide = 1,
            ide = i + 1,
            node_ides = nodal_incidence
         )

      # Escrevendo Dados do .dat
      dati = DAT_Interpreter()
      dati.model = self.model
      self.data = dati.write()

# --------------------------------------------------
# 4 - Classes do Artefato "cuboid"
# --------------------------------------------------
class Cuboid(Artifact):
   # Funções de Geração de Coordenadas e Incidência de Elementos
   def _brick20_coordinates(self):
      # Renomeando Atributos
      width, height, deep = self.dimensions
      nx, ny, nz = self.discretization

      # Calculando Valores Necessários
      delta_x = width / nx
      delta_y = height / ny
      delta_z = deep / nz
      x_values = [delta_x / 2 * i for i in range(2 * nx + 1)]
      y_values = [delta_y / 2 * i for i in range(2 * ny + 1)]
      z_values = [delta_z / 2 * i for i in range(2 * nz + 1)]

      # Gerando Coordenadas
      ide = 1
      for i_z, z in enumerate(z_values):
         # Mudando Valores de y com base em z
         ys = y_values if i_z % 2 == 0 else y_values[::2]

         for i_y, y in enumerate(ys):
            # Mudando Valores de x com base em y e z
            xs = x_values
            if (
               ((i_z % 2 == 0) and (i_y % 2 == 1)) or
               (i_z % 2 == 1)
            ):
               xs = x_values[::2]

            for x in xs:
               self.model.add_node(ide, x, y, z)
               ide += 1

   def _brick20_incidence(self, i: int) -> list[int]:
      # Inicializando Variáveis
      nx, ny, _ = self.discretization
      inc = [0] * 20
      x_order = i % nx
      if x_order == 0: x_order = nx
      layer_order = i % (nx * ny)
      if layer_order == 0: layer_order = nx * ny
      y_order = layer_order // nx
      if layer_order % nx != 0: y_order += 1
      z_order = i // (nx * ny)
      if i % (nx * ny) != 0: z_order += 1
      
      # Determinando Nó Inicial
      inc[0] = (2 * x_order - 1) + (y_order - 1) * (3 * nx + 2) + (z_order - 1) * ((nx + 1) * (ny + 1) + (2 * nx + 1) * (2 * ny + 1) - nx * ny)

      # Camada 1
      inc[1] = inc[0] + 1
      inc[2] = inc[0] + 2
      inc[3] = inc[0] + 2 * nx + 3 - x_order
      inc[4] = inc[3] + nx + 1 + x_order
      inc[5] = inc[4] - 1
      inc[6] = inc[4] - 2
      inc[7] = inc[3] - 1

      # Camada 2
      inc[8] = inc[0] + 2 * nx - x_order + 2 + (ny + 1 - y_order) * (3 * nx + 2) + (y_order - 1) * (nx + 1)
      inc[9] = inc[8] + 1
      inc[10] = inc[8] + nx + 2
      inc[11] = inc[10] - 1
      
      # Camada 3
      inc[12] = inc[8] + nx - x_order + 1 + (ny + 1 - y_order) * (nx + 1) + (y_order - 1) * (3 * nx + 2) + 2 * (x_order - 1) + 1
      inc[13] = inc[12] + 1
      inc[14] = inc[12] + 2
      inc[15] = inc[12] + 2 * nx + 3 - x_order
      inc[16] = inc[15] + nx + 1 + x_order
      inc[17] = inc[16] - 1
      inc[18] = inc[16] - 2
      inc[19] = inc[15] - 1

      # Ordenando Conforme o FAST
      fast_order = [12, 13, 14, 15, 16, 17, 18, 19, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5, 6, 7]
      inc = [inc[i] for i in fast_order]

      return inc

   # Elementos Suportados
   supported_elements = {
      'BRICK20': {
         'coordinates': _brick20_coordinates,
         'incidence': _brick20_incidence
      }
   }

   def __init__(
      self,
      element_type: str,
      dimensions: list[float],
      discretization: list[int]
   ):
      # Chamando Construtor da Superclasse
      super().__init__('cuboid', 'dat')

      # Verificando se Tipo de Elemento Fornecido é Suportado
      if element_type not in Cuboid.supported_elements.keys():
         raise ValueError(f'Element Type "{element_type}" is not supported for cuboid generation.')   

      # Verificando se o Número de Dimensões e Discretização foram passadas Corretamente
      if len(dimensions) != 3:
         raise ValueError('A Cuboid needs exactly 3 dimensions (width, height and deep).')
      if len(discretization) != 3:
         raise ValueError('A Cuboid needs exactly 3 discretization values (number of elements in width, height and deep).')

      # Atribuindo Atributos
      self.element_type = element_type
      self.dimensions = dimensions
      self.discretization = discretization
      self.model = SimulationModel()
      self._coordinates = Cuboid.supported_elements[element_type]['coordinates']
      self._incidence = Cuboid.supported_elements[element_type]['incidence']
      self.reference = searcher.get_database('translation_reference')

   def coordinates(self):
      return self._coordinates(self)
   
   def incidence(self, i: int) -> list[int]:
      return self._incidence(self, i)
   
   def geometry(self) -> int:
      element_info = self.reference['dat']['elements'][self.element_type]
      return self.model.add_element_geometry(**element_info)

   def generate(self):
      # Renomeando Atributos
      nx, ny, nz = self.discretization

      # Gerando Coordenadas e Nodes
      self.coordinates()

      # Configurações dos Elementos
      geometry_ide = self.geometry()
      self.model.add_element_group(1, geometry_ide, None)

      # Gerando Elementos
      for i in range(nx * ny * nz):
         # Gerando Incidência
         nodal_incidence = self.incidence(i + 1)

         # Cadastrando Elemento
         self.model.add_element(
            group_ide = 1,
            ide = i + 1,
            node_ides = nodal_incidence
         )

      # Escrevendo Dados do .dat
      dati = DAT_Interpreter()
      dati.model = self.model
      self.data = dati.write()

# --------------------------------------------------
# 5 - Classes do Artefato "nurbs_rectangle"
# --------------------------------------------------
class NURBS_Rectangle(Artifact):
   def __init__(
      self,
      degrees: list[int],
      dimensions: list[float],
      discretization: list[int]
   ):
      # Chamando Construtor da Superclasse
      super().__init__('nurbs_rectangle', 'dat')  

      # Verificando se o Número de Dimensões e Discretização foram passadas Corretamente
      if len(degrees) != 2:
         raise ValueError('A NURBS Rectangle needs exactly 2 degrees (degrees in width and height).')
      if len(dimensions) != 2:
         raise ValueError('A NURBS Rectangle needs exactly 2 dimensions (width and height).')
      if len(discretization) != 2:
         raise ValueError('A NURBS Rectangle needs exactly 2 discretization values (number of elements in width and height).')

      # Atribuindo Atributos
      self.degrees = degrees
      self.dimensions = dimensions
      self.discretization = discretization
      self.model = SimulationModel()

   def coordinates(self) -> list[NURBS_Curve]:
      # Renomeando Atributos
      dx, dy = self.degrees
      width, height = self.dimensions
      nx, ny = self.discretization

      # Calculando Valores Necessários (Em X)
      nurbs_x = NURBS_Curve(1, [0, 0, 1, 1], [[0.0], [width]], [1.0, 1.0])
      nurbs_x.degree_elevation(dx - 1)
      x_knots = [i / nx for i in range(1, nx)]
      nurbs_x.knot_insertions(x_knots)
      x_values = [cp[0] for cp in nurbs_x.control_points]

      # Calculando Valores Necessários (Em y)
      nurbs_y = NURBS_Curve(1, [0, 0, 1, 1], [[0.0], [height]], [1.0, 1.0])
      nurbs_y.degree_elevation(dy - 1)
      y_knots = [i / ny for i in range(1, ny)]
      nurbs_y.knot_insertions(y_knots)
      y_values = [cp[0] for cp in nurbs_y.control_points]

      # Gerando Coordenadas
      ide = 1
      for y in y_values:
         for x in x_values:
            self.model.add_node(ide, x, y, 0.0, 1.0)
            ide += 1

      return nurbs_x, nurbs_y

   def generate(self):
      # Renomeando Atributos
      dx, dy = self.degrees
      nx, ny = self.discretization

      # Gerando Coordenadas e Nodes
      nurbs_x, nurbs_y = self.coordinates()

      # Configurações dos Elementos
      node_space = list(
         map(
            lambda n: n + 1,
            range(len(self.model.nodes))
         )
      )
      geometry_ide = self.model.add_element_geometry(
         shape = 'Quadrilateral',
         base = 'BSpline',
         grade = [nurbs_x.basis.degree, nurbs_y.basis.degree],
         n_nodes = (nurbs_x.basis.degree + 1) * (nurbs_y.basis.degree + 1),
         n_dimensions = 2,
         knot_vectors = [nurbs_x.basis.knot_vector, nurbs_y.basis.knot_vector],
         node_space = node_space,
      )
      self.model.add_element_group(1, geometry_ide, 'ShallowShell')

      # Gerando Matriz do Node Space
      node_space_matrix = list()
      for i in range(0, len(node_space), nx + dx):
         node_space_matrix.append(node_space[i:i + nx + dx])

      # Gerando Elementos
      ide = 1
      for ks2 in range(ny):
         for ks1 in range(nx):
            # Determinando Nodes do Elemento
            node_ides = list()
            for i in range(ks2, ks2 + dy + 1):
               for j in range(ks1, ks1 + dx + 1):
                  node_ides.append(node_space_matrix[i][j])

            # Cadastrando Elemento
            self.model.add_element(
               group_ide = 1,
               ide = ide,
               node_ides = node_ides,
               knot_span = [ks1 + 1, ks2 + 1]
            )
            ide += 1

      # Escrevendo Dados do .dat
      dati = DAT_Interpreter()
      dati.model = self.model
      self.data = dati.write()
   
# --------------------------------------------------
# 6 - Classes do Artefato "nurbs_cuboid"
# --------------------------------------------------
class NURBS_Cuboid(Artifact):
   def __init__(
      self,
      degrees: list[int],
      dimensions: list[float],
      discretization: list[int]
   ):
      # Chamando Construtor da Superclasse
      super().__init__('nurbs_cuboid', 'dat')  

      # Verificando se o Número de Dimensões e Discretização foram passadas Corretamente
      if len(degrees) != 3:
         raise ValueError('A NURBS Cuboid needs exactly 3 degrees (degrees in width, height and deep).')
      if len(dimensions) != 3:
         raise ValueError('A NURBS Cuboid needs exactly 3 dimensions (width, height and deep).')
      if len(discretization) != 3:
         raise ValueError('A NURBS Cuboid needs exactly 3 discretization values (number of elements in width, height and deep).')

      # Atribuindo Atributos
      self.degrees = degrees
      self.dimensions = dimensions
      self.discretization = discretization
      self.model = SimulationModel()

   def coordinates(self) -> list[NURBS_Curve]:
      # Renomeando Atributos
      dx, dy, dz = self.degrees
      width, height, deep = self.dimensions
      nx, ny, nz = self.discretization

      # Calculando Valores Necessários (Em X)
      nurbs_x = NURBS_Curve(1, [0, 0, 1, 1], [[0.0], [width]], [1.0, 1.0])
      nurbs_x.degree_elevation(dx - 1)
      x_knots = [i / nx for i in range(1, nx)]
      nurbs_x.knot_insertions(x_knots)
      x_values = [cp[0] for cp in nurbs_x.control_points]

      # Calculando Valores Necessários (Em y)
      nurbs_y = NURBS_Curve(1, [0, 0, 1, 1], [[0.0], [height]], [1.0, 1.0])
      nurbs_y.degree_elevation(dy - 1)
      y_knots = [i / ny for i in range(1, ny)]
      nurbs_y.knot_insertions(y_knots)
      y_values = [cp[0] for cp in nurbs_y.control_points]

      # Calculando Valores Necessários (Em z)
      nurbs_z = NURBS_Curve(1, [0, 0, 1, 1], [[0.0], [deep]], [1.0, 1.0])
      nurbs_z.degree_elevation(dz - 1)
      z_knots = [i / nz for i in range(1, nz)]
      nurbs_z.knot_insertions(z_knots)
      z_values = [cp[0] for cp in nurbs_z.control_points]

      # Gerando Coordenadas
      ide = 1
      for z in reversed(z_values):
         for y in y_values:
            for x in x_values:
               self.model.add_node(ide, x, y, z, 1.0)
               ide += 1

      return nurbs_x, nurbs_y, nurbs_z

   def generate(self):
      # Renomeando Atributos
      dx, dy, dz = self.degrees
      nx, ny, nz = self.discretization

      # Gerando Coordenadas e Nodes
      nurbs_x, nurbs_y, nurbs_z = self.coordinates()

      # Configurações dos Elementos
      node_space = list(
         map(
            lambda n: n + 1,
            range(len(self.model.nodes))
         )
      )
      geometry_ide = self.model.add_element_geometry(
         shape = 'Hexahedron',
         base = 'BSpline',
         grade = [nurbs_x.basis.degree, nurbs_y.basis.degree, nurbs_z.basis.degree],
         n_nodes = (nurbs_x.basis.degree + 1) * (nurbs_y.basis.degree + 1) * (nurbs_z.basis.degree + 1),
         n_dimensions = 3,
         knot_vectors = [nurbs_x.basis.knot_vector, nurbs_y.basis.knot_vector, nurbs_z.basis.knot_vector],
         node_space = node_space,
      )
      self.model.add_element_group(1, geometry_ide, None)

      # Gerando Matriz do Node Space
      node_space_matrix = list()
      for i in range(0, len(node_space), (ny + dy) * (nx + dx)):
         node_space_matrix.append(list())
         for j in range(i, i + (ny + dy) * (nx + dx), nx + dx):
            node_space_matrix[-1].append(node_space[j:j + nx + dx])

      # Gerando Elementos
      ide = 1
      for ks3 in range(nz):
         for ks2 in range(ny):
            for ks1 in range(nx):
               # Determinando Nodes do Elemento
               node_ides = list()
               for i in range(ks3, ks3 + dz + 1):
                  for j in range(ks2, ks2 + dy + 1):
                     for l in range(ks1, ks1 + dx + 1):
                        node_ides.append(node_space_matrix[i][j][l])

               # Cadastrando Elemento
               self.model.add_element(
                  group_ide = 1,
                  ide = ide,
                  node_ides = node_ides,
                  knot_span = [ks1 + 1, ks2 + 1, ks3 + 1]
               )
               ide += 1

      # Escrevendo Dados do .dat
      dati = DAT_Interpreter()
      dati.model = self.model
      self.data = dati.write()

# --------------------------------------------------
# 6 - Classes do Artefato "cyl_panel"
# --------------------------------------------------
class CylindricalPanel(Artifact):
      # Funções de Geração de Coordenadas e Incidência de Elementos
   def _q8_coordinates(self):
      # Renomeando Atributos
      height = self.height
      nx, ny = self.discretization
      R = self.radius
      a1, a2 = self.angles

      # Calculando Valores Necessários (Espaço Paramétrico)
      delta_x = 1.0 / nx
      delta_y = 1.0 / ny
      x_values = [delta_x / 2 * i for i in range(2 * nx + 1)]
      y_values = [delta_y / 2 * i for i in range(2 * ny + 1)]

      # Gerando Coordenadas
      ide = 1
      for i_y, y in enumerate(y_values):
         # Mudando Valores de x com base em y
         xs = x_values if i_y % 2 == 0 else x_values[::2]

         for x in xs:
            # Convertendo Coordenadas do Espaço Paramétrico para o Real
            theta = a1 + x * (a2 - a1)
            x_r = R * cos(theta / 180 * pi)
            z_r = R * sin(theta / 180 * pi)
            y_r = y * height

            self.model.add_node(ide, x_r, y_r, z_r)
            ide += 1

   def _q8_incidence(self, i: int) -> list[int]:
      # Inicializando Variáveis
      nx, _ = self.discretization
      inc = [0] * 8
      x_order = i % nx
      if x_order == 0: x_order = nx
      y_order = i // nx
      if i % nx != 0: y_order += 1
      
      # Determinando Nó Inicial
      inc[0] = (2 * x_order - 1) + (y_order - 1) * (3 * nx + 2)

      # Determinando Restante da Incidência
      inc[1] = inc[0] + 1
      inc[2] = inc[0] + 2
      inc[3] = inc[0] + 2 * nx + 3 - x_order
      inc[4] = inc[3] + nx + 1 + x_order
      inc[5] = inc[4] - 1
      inc[6] = inc[4] - 2
      inc[7] = inc[3] - 1

      # Ordenando Conforme o FAST
      fast_order = [0, 1, 2, 3, 4, 5, 6, 7]
      inc = [inc[i] for i in fast_order]

      return inc

   # Elementos Suportados
   supported_elements = {
      'Q8': {
         'coordinates': _q8_coordinates,
         'incidence': _q8_incidence
      }
   }

   def __init__(
      self,
      element_type: str,
      height: float,
      discretization: list[int],
      radius: float,
      angles: list[float]
   ):
      # Chamando Construtor da Superclasse
      super().__init__('cyl_panel', 'dat')

      # Verificando se Tipo de Elemento Fornecido é Suportado
      if element_type not in CylindricalPanel.supported_elements.keys():
         raise ValueError(f'Element Type "{element_type}" is not supported for cylindrical panel generation.')   

      # Verificando se os Parâmetros foram passados Corretamente
      if len(discretization) != 2:
         raise ValueError('A Cylindrical Panel needs exactly 2 discretization values (number of elements in width and height).')
      if len(angles) != 2:
         raise ValueError('A Cylindrical Panel needs exactly 2 angle values (start angle and stop angle).')

      # Atribuindo Atributos
      self.element_type = element_type
      self.height = height
      self.discretization = discretization
      self.radius = radius
      self.angles = angles
      self.model = SimulationModel()
      self._coordinates = CylindricalPanel.supported_elements[element_type]['coordinates']
      self._incidence = CylindricalPanel.supported_elements[element_type]['incidence']
      self.reference = searcher.get_database('translation_reference')

   def coordinates(self):
      return self._coordinates(self)
   
   def incidence(self, i: int) -> list[int]:
      return self._incidence(self, i)
   
   def geometry(self) -> int:
      element_info = self.reference['dat']['elements'][self.element_type]
      return self.model.add_element_geometry(**element_info)

   def generate(self):
      # Renomeando Atributos
      nx, ny = self.discretization

      # Gerando Coordenadas e Nodes
      self.coordinates()

      # Configurações dos Elementos
      geometry_ide = self.geometry()
      self.model.add_element_group(1, geometry_ide, None)

      # Gerando Elementos
      for i in range(nx * ny):
         # Gerando Incidência
         nodal_incidence = self.incidence(i + 1)

         # Cadastrando Elemento
         self.model.add_element(
            group_ide = 1,
            ide = i + 1,
            node_ides = nodal_incidence
         )

      # Escrevendo Dados do .dat
      dati = DAT_Interpreter()
      dati.model = self.model
      self.data = dati.write()

# --------------------------------------------------
# 6 - Classes do Artefato "slit_annular_plate"
# --------------------------------------------------
class SlitAnnularPlate(Artifact):
      # Funções de Geração de Coordenadas e Incidência de Elementos
   def _q8_coordinates(self):
      # Renomeando Atributos
      nx, ny = self.discretization
      R_i = self.inner_radius
      R_o = self.outer_radius

      # Calculando Valores Necessários (Espaço Paramétrico)
      delta_x = 1.0 / nx
      delta_y = 1.0 / ny
      x_values = [delta_x / 2 * i for i in range(2 * nx + 1)]
      y_values = [delta_y / 2 * i for i in range(2 * ny + 1)]

      # Gerando Coordenadas
      ide = 1
      for i_y, y in enumerate(y_values):
         # Mudando Valores de x com base em y
         xs = x_values if i_y % 2 == 0 else x_values[::2]

         for x in xs:
            # Convertendo Coordenadas do Espaço Paramétrico para o Real
            R = R_i + (R_o - R_i) * x
            theta = 2 * pi * y

            x_r = R * cos(theta)
            y_r = R * sin(theta)

            self.model.add_node(ide, x_r, y_r, 0.0)
            ide += 1

   def _q8_incidence(self, i: int) -> list[int]:
      # Inicializando Variáveis
      nx, _ = self.discretization
      inc = [0] * 8
      x_order = i % nx
      if x_order == 0: x_order = nx
      y_order = i // nx
      if i % nx != 0: y_order += 1
      
      # Determinando Nó Inicial
      inc[0] = (2 * x_order - 1) + (y_order - 1) * (3 * nx + 2)

      # Determinando Restante da Incidência
      inc[1] = inc[0] + 1
      inc[2] = inc[0] + 2
      inc[3] = inc[0] + 2 * nx + 3 - x_order
      inc[4] = inc[3] + nx + 1 + x_order
      inc[5] = inc[4] - 1
      inc[6] = inc[4] - 2
      inc[7] = inc[3] - 1

      # Ordenando Conforme o FAST
      fast_order = [0, 1, 2, 3, 4, 5, 6, 7]
      inc = [inc[i] for i in fast_order]

      return inc

   # Elementos Suportados
   supported_elements = {
      'Q8': {
         'coordinates': _q8_coordinates,
         'incidence': _q8_incidence
      }
   }

   def __init__(
      self,
      element_type: str,
      inner_radius: float,
      outer_radius: float,
      discretization: list[int]
   ):
      # Chamando Construtor da Superclasse
      super().__init__('slit_annular_plate', 'dat')

      # Verificando se Tipo de Elemento Fornecido é Suportado
      if element_type not in CylindricalPanel.supported_elements.keys():
         raise ValueError(f'Element Type "{element_type}" is not supported for slit annular plate generation.')   

      # Verificando se os Parâmetros foram passados Corretamente
      if len(discretization) != 2:
         raise ValueError('A Slit Annular Plate needs exactly 2 discretization values (number of elements along radius and circumference).')
      
      # Atribuindo Atributos
      self.element_type = element_type
      self.inner_radius = inner_radius
      self.outer_radius = outer_radius
      self.discretization = discretization
      self.model = SimulationModel()
      self._coordinates = SlitAnnularPlate.supported_elements[element_type]['coordinates']
      self._incidence = SlitAnnularPlate.supported_elements[element_type]['incidence']
      self.reference = searcher.get_database('translation_reference')

   def coordinates(self):
      return self._coordinates(self)
   
   def incidence(self, i: int) -> list[int]:
      return self._incidence(self, i)
   
   def geometry(self) -> int:
      element_info = self.reference['dat']['elements'][self.element_type]
      return self.model.add_element_geometry(**element_info)

   def generate(self):
      # Renomeando Atributos
      nx, ny = self.discretization

      # Gerando Coordenadas e Nodes
      self.coordinates()

      # Configurações dos Elementos
      geometry_ide = self.geometry()
      self.model.add_element_group(1, geometry_ide, None)

      # Gerando Elementos
      for i in range(nx * ny):
         # Gerando Incidência
         nodal_incidence = self.incidence(i + 1)

         # Cadastrando Elemento
         self.model.add_element(
            group_ide = 1,
            ide = i + 1,
            node_ides = nodal_incidence
         )

      # Escrevendo Dados do .dat
      dati = DAT_Interpreter()
      dati.model = self.model
      self.data = dati.write()


# --------------------------------------------------
# 1 - Classes do Modelo de Simulação
# --------------------------------------------------
class SimulationModel:
   def __init__(self):
      # Nodes - Atributos Relacionados
      self.nodes = dict[int, Node]()
      self.node_sets = dict[str, set]()
      self.node_solver_order = list[int]()

      # Elements - Atributos Relacionados
      self.element_geometries = dict[int, ElementGeometry]()
      self.element_groups = dict[int, ElementGroup]()
      self.element_sets = dict[str, set]()

      # Supports - Atributos Relacionados
      self.supports = dict[int, set]()
      self.supported_dofs = ('u', 'v', 'w', 'rx', 'ry', 'rz')

      # Materials - Atributos Relacionados
      self.materials = dict[int, Material]()

      # Sections - Atributos Relacionados
      self.sections = dict[int, Section]()
   
   # Métodos - Adição de Entidades
   def add_node(
      self, 
      ide: int, 
      x: float, 
      y: float, 
      z: float, 
      weight: float = None
   ):
      self.nodes[ide] = Node(x, y, z, weight)
   
   def add_element_geometry(
      self,
      shape: str,
      base: str,
      grade: int | list[int],
      n_nodes: int,
      n_dimensions: int,
      knot_vectors: list[list[float]] = None, 
      node_space: list[int] = None
   ) -> int:
      # Verificando se Geometria Já Existe
      for geometry_ide, element_geometry in self.element_geometries.items():
         if (
            (shape == element_geometry.shape) and
            (base == element_geometry.base) and
            (grade == element_geometry.grade) and
            (n_nodes == element_geometry.n_nodes) and
            (n_dimensions == element_geometry.n_dimensions)
         ):
            if base == 'BSpline':
               if knot_vectors == element_geometry.knot_vectors and node_space == element_geometry.node_space:
                  break
               continue
            break
      else:
         # Criando Geometria (Já que não Existe)
         geometry_ide = len(self.element_geometries) + 1
         self.element_geometries[geometry_ide] = ElementGeometry(shape, base, grade, n_nodes, n_dimensions, knot_vectors, node_space)

      # Retornando Ide da Geometria
      return geometry_ide
   
   def add_element_group(self, ide: int, geometry_ide, theory: str):
      if geometry_ide not in self.element_geometries:
         raise ValueError(f'The Element Geometry with ide = {geometry_ide} does not exist.')
      self.element_groups[ide] = ElementGroup(geometry_ide, theory)

   def add_element(
      self, group_ide: int,
      ide: int,
      node_ides: list[int], 
      knot_span: list[int] = None
   ):
      # Verificando se Ides de Nodes são Válidos
      for node_ide in node_ides:
         if node_ide not in self.nodes:
            raise ValueError(f'The Node with ide = {node_ide} does not exist.')
      
      # Criando Elemento
      self.element_groups[group_ide].elements[ide] = Element(node_ides, knot_span)
   
   def add_support(self, node_ide: int, dof: str):
      # Verificando Entradas
      if node_ide not in self.nodes:
         raise ValueError(f'The Node with ide = {node_ide} does not exist.')
      if dof not in self.supported_dofs:
         raise ValueError(f'The Degree of Freedom "{dof}" is not supported.')
      
      # Relacionando Grau de Liberdade Restrito com o Node
      if self.supports.get(node_ide) is None:
         self.supports[node_ide] = set()
      self.supports[node_ide].add(dof)

# --------------------------------------------------
# 2 - Classes Relacionadas à Malha
# --------------------------------------------------
class Node:
   def __init__(self, x: float, y: float, z: float, weight: float = None):
      self.x = x
      self.y = y
      self.z = z
      self.weight = weight

class ElementGeometry:
   def __init__(
      self,
      shape: str, 
      base: str, 
      grade: int | list[int],
      n_nodes: int,
      n_dimensions: int,
      knot_vectors: list[list[float]] = None, 
      node_space: list[int] = None
   ):
      # Atributos de Geometrias em Geral
      self.shape = shape
      self.base = base
      self.grade = grade
      self.n_nodes = n_nodes
      self.n_dimensions = n_dimensions

      # Atributos de Geometria com Base BSpline
      self.knot_vectors = knot_vectors
      self.node_space = node_space

class ElementGroup:
   def __init__(self, geometry_ide: int, theory: str):
      self.geometry_ide = geometry_ide
      self.theory = theory
      self.elements = dict[int, Element]()

class Element:
   def __init__(self, node_ides: list[int], knot_span: list[int] = None):
      self.node_ides = node_ides
      self.knot_span = knot_span

# --------------------------------------------------
# 3 - Classes Relacionadas a Materiais
# --------------------------------------------------
class Material:
   pass

class IsotropicMaterial(Material):
   def __init__(self, elastic_modulus: float, poisson_coefficient: float, density: float) -> None:
      self.E = elastic_modulus
      self.nu = poisson_coefficient
      self.rho = density

      # Calculando Módulo Volumétrico
      self.K = self.E / (3 * (1 - 2 * self.nu))

      # Calculando Módulo de Cisalhamento
      self.G = self.E / (2 * (1 + self.nu))

class FunctionallyGradedMaterial(Material):
   # Funções de Homogeneização Privadas
   def _voigt(self, volume_fractions: list[float]):
      E, nu, rho = 0, 0, 0
      for V, M in zip(volume_fractions, self.materials):
         E += V * M.E
         nu += V * M.nu
         rho += V * M.rho
      return E, nu, rho
   
   def _hashin_shtrikman(bound: str):
      def function(self, volume_fractions: list[float]) -> list[float]:
         # Definindo Valores Especiais
         V, M = volume_fractions, self.materials
         V1, V2 = V[0], V[1]
         K1, K2 = M[0].K, M[1].K
         G1, G2 = M[0].G, M[1].G
         rho_1, rho_2 = M[0].rho, M[1].rho

         # Valores Iniciais do que é Matriz e do que são as Inclusões
         Vm, Vi = V1, V2
         Km, Ki = K1, K2
         Gm, Gi = G1, G2
         rho_m, rho_i = rho_1, rho_2

         # Trocando Ordem com base no Bound Escolhido
         if (
            bound == 'upper' and M[0].E < M[1].E or
            bound == 'lower' and M[0].E > M[1].E
         ):
            Vm, Vi = V2, V1
            Km, Ki = K2, K1
            Gm, Gi = G2, G1
            rho_m, rho_i = rho_2, rho_1

         # Calculando Valores Auxilizares
         FK = (3 * Vm) / (3 * Km + 4 * Gm)
         FG = 6 * Vm * (Km + 2 * Gm) / (5 * Gm * (3 * Km + 4 * Gm))

         # Calculando Módulo Volumétrico
         K = Km + Vi / ((1 / (Ki - Km)) + FK)

         # Calculando Módulo de Cisalhamento
         G = Gm + Vi / ((1 / (Gi - Gm)) + FG)

         # Calculando Propriedades Efetivas
         E = (9 * G * K) / (G + 3 * K)
         nu = (3 * K - 2 * G) / (2 * (G + 3 * K))

         # Densidade Calculada pelo Modelo de voigt
         rho = Vi * rho_i + Vm * rho_m

         return E, nu, rho
      return function

   # Relação Modelo/Função de Homogeneização
   homogenize_functions = {
      'voigt': _voigt,
      'mori_tanaka': _hashin_shtrikman('lower'),
      'hashin_shtrikman_upper_bound': _hashin_shtrikman('upper'),
      'hashin_shtrikman_lower_bound': _hashin_shtrikman('lower'),
   }

   def __init__(self, micromechanical_model: str, materials: list[IsotropicMaterial]) -> None:
      self.micromechanical_model = micromechanical_model
      self.materials = materials
      try:
         self._homogenize = FunctionallyGradedMaterial.homogenize_functions[micromechanical_model]
      except KeyError:
         raise ValueError(f'Micromechanical Model "{micromechanical_model}" is not supported.')
   
   def homogenize(self, volume_fractions: list[float]):
      return self._homogenize(self, volume_fractions)

# --------------------------------------------------
# 4 - Classes Relacionadas a Seções
# --------------------------------------------------
class Section:
   def __init__(self, material_ide: int) -> None:
      self.material_ide = material_ide

class HIM_3D_Section(Section):
   def __init__(self, material_ide: int) -> None:
      # Chamando Construtor da Superclasse
      super().__init__(material_ide)

class FGM_3D_Section(Section):
   def __init__(self, material_ide: int, volume_fractions: dict[int, float]) -> None:
      # Chamando Construtor da Superclasse
      super().__init__(material_ide)

      # Atribuindo Atributos
      self.volume_fractions = volume_fractions


import re
from math import floor
from ..interface import searcher
from .geometry import (
   bezier_equiv_coord, 
   bernstein_polynomial,
   NURBS_Surface,
   GeometricalTransformer
)
from .simulation import (
   SimulationModel,
   Node,
   ElementGeometry,
   ElementGroup,
   IsotropicMaterial,
   FunctionallyGradedMaterial,
   HIM_3D_Section,
   FGM_3D_Section
)

class INP_Interpreter:
   def __init__(self):
      self.model = SimulationModel()
      self.reference = searcher.get_database('translation_reference')['inp']
   
   def read_nodes(self, inp_data: str):
      # Identificando Nodes
      keyword_format = '\*Node\n([^*]*)'
      node = '(-?\d+.\d*e?-?\+?\d*)'
      line_format = f'(\d+),\s*{node},\s*{node}(?:,\s*{node})?'

      # Inserindo Nodes
      lines_data = re.findall(keyword_format, inp_data)[0]
      nodes = re.findall(line_format, lines_data)
      for node in nodes:
         ide, x, y, z = node

         # Convertendo Valores
         ide = int(ide)
         x, y = float(x), float(y)
         z = 0.0 if z == '' else float(z)

         self.model.add_node(ide, x, y, z)
   
   def read_node_sets(self, inp_data: str):
      # Identificando Conjuntos de Nodes
      keyword_format = '\*Nset\s*,\s*nset=([^,\n]+)([^*]+)'
      sets_data = re.findall(keyword_format, inp_data)

      # Tratando Informações
      for set_data in sets_data:
         # Nomeando Informações
         set_name = set_data[0]
         first_break_line_index = set_data[1].index('\n')
         set_params = set_data[1][:first_break_line_index]
         set_numbers = set_data[1][first_break_line_index + 1:]
         
         # Obtendo Números (Ides de Nodes ou Parâmetros de Range)
         number_format = '(\d+)'
         numbers = re.findall(number_format, set_numbers)
         numbers = list(map(int, numbers))

         # Tratando de Acordo com o tipo do Conjunto
         if 'generate' in set_params:
            numbers[1] += 1
            self.model.node_sets[set_name] = range(*numbers)
         else:
            self.model.node_sets[set_name] = numbers
   
   def read_elements(self, inp_data: str):
      # Identificando Grupos de Elementos
      keyword_format = '\*Element, type=(.*)\n([^*]*)'
      groups_data = re.findall(keyword_format, inp_data)

      # Analisando Cada Grupo
      group_ide = 1
      for element_type, lines_data in groups_data:
         # Identificando Elementos
         try:
            type_info = self.reference['elements'][element_type]
         except KeyError:
            raise KeyError(f'The Element Type "{element_type}" is not supported  for .inp files.')
         int_ide = '(\d+)'
         node_ide = ',\s*' + int_ide
         line_format = int_ide + type_info['n_nodes'] * node_ide
         elements = re.findall(line_format, lines_data)

         # Criando Geometria
         geometry_ide = self.model.add_element_geometry(
            type_info['shape'],
            type_info['base'],
            type_info['grade'],
            type_info['n_nodes'],
            type_info['n_dimensions']
         )

         # Criando Grupo de Elementos
         self.model.add_element_group(group_ide, geometry_ide, type_info.get('theory'))

         # Inserindo Elementos
         for element in elements:
            ide, *node_ides = map(int, element)
            self.model.add_element(group_ide, ide, node_ides)
         
         # Incrementando Ide do Grupo
         group_ide += 1
   
   def read_element_sets(self, inp_data: str):
      # Identificando Conjuntos de Elementos
      keyword_format = '\*Elset\s*,\s*elset=([^,\n]+)([^*]+)'
      sets_data = re.findall(keyword_format, inp_data)

      # Tratando Informações
      for set_data in sets_data:
         # Nomeando Informações
         set_name = set_data[0]
         set_params, set_numbers, *_ = set_data[1].split('\n')
         
         # Obtendo Números (Ides de Nodes ou Parâmetros de Range)
         number_format = '(\d+)\s*,?'
         numbers = re.findall(number_format, set_numbers)
         numbers = list(map(int, numbers))

         # Tratando de Acordo com o tipo do Conjunto
         if 'generate' in set_params:
            self.model.element_sets[set_name] = range(*numbers)
         else:
            self.model.element_sets[set_name] = numbers

   def read_supports(self, inp_data: str):
      # Identificando Supports
      keyword_format = '\*Boundary([^*]+)'
      supports_data = re.findall(keyword_format, inp_data)

      # Tratando Informações
      for support_data in supports_data:
         # Obtendo Informações
         line_format = '(\S+)\s*,\s*(\d+|\w+)\s*(?:,\s*(\d+))?'
         lines_data = re.findall(line_format, support_data)

         for line_data in lines_data:
            # Verificando Natureza do Alvo da Condição de Support
            try:
               # O Alvo é um Node
               boundary_target = int(line_data[0])
            except ValueError:
               # O Alvo é um Node Set
               boundary_target = line_data[0]

            # Verificando Natureza do Início da Condição de Support
            try:
               # O Início é um Índice
               boundary_start = int(line_data[1]) - 1
            except ValueError:
               # O Início é um Tipo
               boundary_start = line_data[1]
            
            # Determinando Índices de Condição com Base no Alvo
            if (type(boundary_start) is int):
               # Verificando se há um Índice de Fim
               if line_data[2]:
                  boundary_end = int(line_data[2]) - 1
               else:
                  boundary_end = boundary_start
               indexes = range(boundary_start, boundary_end + 1)
            else:
               # Tentando Identificar Tipo de Condição
               try:
                  indexes = self.reference['boundary_types'][boundary_start]
               except KeyError:
                  raise KeyError(f'The Boundary Type "{boundary_start}" is not supported.')

            # Tratando Adição de Supports de acordo com o Alvo
            if type(boundary_target) is int:
               for index in indexes:
                  self.model.add_support(boundary_target, self.model.supported_dofs[index])
            else:
               for node_ide in self.model.node_sets[boundary_target]:
                  for index in indexes:
                     self.model.add_support(node_ide, self.model.supported_dofs[index])

   def read(self, inp_data: str):
      # Interpretando Nodes
      self.read_nodes(inp_data)
      self.read_node_sets(inp_data)

      # Interpretando Elementos
      self.read_elements(inp_data)
      self.read_element_sets(inp_data)

      # Interpretando Supports
      self.read_supports(inp_data)

class DAT_Interpreter:
   # Funções Privadas de Adição de Elementos Específicos
   def _add_bezier_triangles(self, group_ide: int, lines_data: str, element_theory: str):
      # Identificando Elementos
      int_ide = '(\d+)'
      property_ides = '\s+\d+' * 3 + '\s+(\d+)'
      line_format = int_ide + property_ides + '\s+(.+)'
      elements = re.findall(line_format, lines_data)
      
      # Ides de Grupos Relacionados com o Grau dos Elementos
      grade_to_group = dict()

      # Analisando Cada Elemento
      for ide, grade, node_ides in elements:
         # Tipificando Valores
         ide = int(ide)
         grade = int(grade)
         node_ides = list(map(int, node_ides.split()))

         # Verificando se Grupo com o Grau do Elemento Já Existe
         if grade not in grade_to_group:
            grade_to_group[grade] = group_ide
            geometry_ide = self.model.add_element_geometry(
               shape = 'Triangle',
               base = 'Bezier',
               grade = grade,
               n_nodes = len(node_ides),
               n_dimensions = 2
            )
            self.model.add_element_group(group_ide, geometry_ide, element_theory)
            group_ide += 1

         # Inserindo Elementos
         self.model.add_element(grade_to_group[grade], ide, node_ides)
      
      # Retornando Último Valor de Ide de Grupo
      return group_ide

   def _add_bezier_surface(self, group_ide: int, lines_data: str, element_theory: str):
      # Identificando Elementos
      int_ide = '(\d+)'
      property_ides = '\s+\d+' * 3 + '\s+(\d+)' * 2
      line_format = int_ide + property_ides + '\s+(.+)'
      elements = re.findall(line_format, lines_data)
      
      # Ides de Grupos Relacionados com o Grau dos Elementos
      grade_to_group = dict()

      # Analisando Cada Elemento
      for ide, grade_1, grade_2, node_ides in elements:
         # Tipificando Valores
         ide = int(ide)
         grade_1 = int(grade_1)
         grade_2 = int(grade_2)
         grade = (grade_1, grade_2)
         node_ides = list(map(int, node_ides.split()))

         # Verificando se Grupo com o Grau do Elemento Já Existe
         if grade not in grade_to_group:
            grade_to_group[grade] = group_ide
            geometry_ide = self.model.add_element_geometry(
               shape = 'Quadrilateral',
               base = 'Bezier',
               grade = grade,
               n_nodes = len(node_ides),
               n_dimensions = 2
            )
            self.model.add_element_group(group_ide, geometry_ide, element_theory)
            group_ide += 1

         # Inserindo Elementos
         self.model.add_element(grade_to_group[grade], ide, node_ides)
      
      # Retornando Último Valor de Ide de Grupo
      return group_ide
   
   def _add_bspline_elements(self, group_ide: int, lines_data: str, element_theory: str):
      # Identificando Elementos
      int_ide = '(\d+)'
      property_ides = '\s+\d+' * 2 + '\s+(\d+)'
      line_format = int_ide + property_ides + '\s+(.+)'
      elements = re.findall(line_format, lines_data)
      
      # Ides de Grupos Relacionados com o Grau dos Elementos
      geometry_to_info = dict()

      # Analisando Cada Elemento
      for ide, geometry_ide, knot_span in elements:
         # Tipificando Valores
         ide = int(ide)
         geometry_ide = int(geometry_ide)
         knot_span = list(map(int, knot_span.split()))

         # Recuperando Geometria
         geometry = self.model.element_geometries[geometry_ide]

         # Verificando se Ide do Patch Já foi Usado
         if geometry_ide not in geometry_to_info:
            # Gerando Matrix do Node Space
            node_space_matrix = list()
            n = [len(kv) - p - 1 for kv, p in zip(geometry.knot_vectors, geometry.grade)]
            if geometry.n_dimensions == 2:
               for i in range(0, len(geometry.node_space), n[0]):
                  node_space_matrix.append(geometry.node_space[i:i + n[0]])
            else:
               for i in range(0, len(geometry.node_space), (n[1]) * (n[0])):
                  node_space_matrix.append(list())
                  for j in range(i, i + (n[1]) * (n[0]), n[0]):
                     node_space_matrix[-1].append(geometry.node_space[j:j + n[0]])

            # Calculando Multiplicidade dos Knots
            unit_knots = [
               list(sorted(set(kv)))
               for kv in geometry.knot_vectors
            ]
            
            # Adicionando Informações Pertinentes
            geometry_to_info[geometry_ide] = {
               'group_ide': group_ide,
               'node_space_matrix': node_space_matrix,
               'unit_knots': unit_knots
            }
            self.model.add_element_group(group_ide, geometry_ide, element_theory)
            group_ide += 1

         # Recuperando Matriz do Node Space e Multiplicidade dos Knots
         unit_knots = geometry_to_info[geometry_ide]['unit_knots']
         node_space_matrix = geometry_to_info[geometry_ide]['node_space_matrix']
         node_ides = list()

         # Renomeando Informações para Simplificar
         kvs = geometry.knot_vectors
         ks = knot_span
         ps = geometry.grade
         us = unit_knots

         # Calculando Pontos de Paradas dos Ranges da Node Space Matrix
         stop = list()
         for kv, k, u in zip(kvs, ks, us):
            stop.append(kv.index(u[k]))
         
         # Tratamento para Elementos 2D
         if geometry.n_dimensions == 2:
            for i in range(stop[1] - ps[1] - 1, stop[1]):
               for j in range(stop[0] - ps[0] - 1, stop[0]):
                  node_ides.append(node_space_matrix[i][j])

         # Tratamento para Elementos 3D
         else:
            for i in range(stop[2] - ps[2] - 1, stop[2]):
               for j in range(stop[1] - ps[1] - 1, stop[1]):
                  for l in range(stop[0] - ps[0] - 1, stop[0]):
                     node_ides.append(node_space_matrix[i][j][l])

         # Inserindo Elementos
         self.model.add_element(
            geometry_to_info[geometry_ide]['group_ide'], 
            ide,
            node_ides,
            knot_span
         )
      
      # Retornando Último Valor de Ide de Grupo
      return group_ide

   # Funções Privadas de Leitura de Material
   def _read_material_isotropic(self, line_data: str) -> None:
      line_data = line_data.split()
      for i in range(0, len(line_data), 3):
         # Separando Valores
         try:
            material_ide, E, nu = line_data[i:i + 3]
         except ValueError:
            raise ValueError('An Isotropic Material does not have the necessary data in .dat file definition.')

         # Convertendo Valores
         material_ide = int(material_ide)
         E = float(E)
         nu = float(nu)

         # Cadastrando Material
         material = IsotropicMaterial(E, nu, 0.0)
         self.model.materials[material_ide] = material

   def _read_material_density(self, line_data: str) -> None:
      line_data = line_data.split()
      for i in range(0, len(line_data), 2):
         # Separando Valores
         try:
            material_ide, rho = line_data[i:i + 2]
         except ValueError:
            raise ValueError('A Density definition of Isotropic Material is not correct in .dat file.')
         
         # Convertendo Valores
         material_ide = int(material_ide)
         rho = float(rho)

         # Verificando se Há um Material Istrópico para Atribuir a Densidade
         if not self.model.materials.get(material_ide):
            raise IndexError(f'The Material with ID "{material_ide}" does not exist.')
         elif not isinstance(self.model.materials[material_ide], IsotropicMaterial):
            raise TypeError(f'The Material with ID "{material_ide}" is not an Isotropic Material.')
         else:
            self.model.materials[material_ide].rho = rho
   
   def _read_material_fgm(self, line_data: str) -> None:
      line_data = line_data.split()
      for i in range(0, len(line_data), 8):
         # Separando Valores
         try:
            material_ide, E1, nu1, rho1, E2, nu2, rho2, mm = line_data[i:i + 8]
         except ValueError:
            raise ValueError('A FGM Material does not have necessary data in .dat file definition.')
         
         # Convertendo Valores
         material_ide = int(material_ide)
         E1, E2 = float(E1), float(E2)
         nu1, nu2 = float(nu1), float(nu2)
         rho1, rho2 = float(rho1), float(rho2)
         mm = int(mm)

         # Verificando se Há Suporte Para o Modelo Micromecânico Fornecido
         mm_relation = DAT_Interpreter.supported_materials['FGM']['mm_relation']
         try:
            mm = mm_relation[mm]
         except KeyError:
            str_supported = ', '.join(f'{n} (\'{m}\')' for n, m in mm_relation.items())
            raise ValueError(f'The FGM Micromechanical Model "{mm}" is not supported, only: {str_supported}.')

         # Cadastrando Material
         m1 = IsotropicMaterial(E1, nu1, rho1)
         m2 = IsotropicMaterial(E2, nu2, rho2)
         material = FunctionallyGradedMaterial(mm, [m1, m2])
         self.model.materials[material_ide] = material

   # Funções Privadas de Escrita de Material
   def _write_material_isotropic(group: dict[int, IsotropicMaterial]) -> str:
      # Iniciando Output
      n_materials = len(group)
      span = len(str(max(group.keys())))
      output = f'\n%MATERIAL.ISOTROPIC\n{n_materials}\n'

      # Escrevendo Cada Material
      for ide, material in group.items():
         output += f'{ide:<{span}}   {material.E:.8e}   {material.nu:.3f}\n'

      # Escrevendo Densidades
      output += f'\n%MATERIAL.DENSITY\n{n_materials}\n'
      for ide, material in group.items():
         output += f'{ide:<{span}}   {material.rho:.8e}\n'
      
      return output

   def _write_material_fgm(group: dict[int, FunctionallyGradedMaterial]) -> str:
      # Iniciando Output
      n_materials = len(group)
      span = len(str(max(group.keys())))
      output = f'\n%MATERIAL.FGM\n{n_materials}\n'

      # Escrevendo Cada Material
      mm_relation = DAT_Interpreter.supported_materials['FGM']['mm_relation']
      for ide, material in group.items():
         # Tentando Determinar o Número do Modelo Micromecânico
         for n, mm_name in mm_relation.items():
            if mm_name == material.micromechanical_model:
               mm_number = n
               break
         else:
            raise ValueError(f'The FGM Micromechanical Model "{material.micromechanical_model}" of Material with ID "{ide}" is not supported for .dat files.')

         # Segregando Materiais
         m1, m2 = material.materials

         # Escrevendo Material
         output += f'{ide:<{span}}   {m1.E:.8e}   {m1.nu:.3f}   {m1.rho:.8e}   {m2.E:.8e}   {m2.nu:.3f}   {m2.rho:.8e}   {mm_number}\n'
      
      return output

   # Materiais Suportados
   supported_materials = {
      'ISOTROPIC': {
         'class': IsotropicMaterial,
         'read': _read_material_isotropic,
         'write': _write_material_isotropic
      },
      'DENSITY': {
         'read': _read_material_density
      },
      'FGM': {
         'class': FunctionallyGradedMaterial,
         'read': _read_material_fgm,
         'write': _write_material_fgm,
         'mm_relation': {
            1: 'voigt',
            2: 'mori_tanaka'
         }
      }
   }

   # Funções Privadas de Leitura de Seção
   def _read_section_him_3d(self, line_data: str) -> None:
      line_data = line_data.split()
      for i in range(0, len(line_data), 2):
         # Separando Valores
         try:
            section_ide, material_ide = map(int, line_data[i:i + 2])
         except ValueError:
            raise ValueError('A Homogeneous Isotropic 3D Section does not have the necessary data in .dat file definition.')

         # Cadastrando Seção
         section = HIM_3D_Section(material_ide)
         self.model.sections[section_ide] = section

   def _read_section_fgm_3d(self, line_data: str) -> None:
      line_data = line_data.split()
      n_volume_fractions = 0
      i = 0
      while i < len(line_data):
         # Separando Valores
         try:
            # Lendo Dados Conhecidos
            section_ide, material_ide, n_volume_fractions = map(int, line_data[i:i + 3])
            
            # Calculando o Index de Início da Próxima Seção
            next_section_index = i + 3 + 2 * n_volume_fractions
            
            # Lendo ID dos Nós e sua Respectiva Fração de Volume
            volume_fractions = dict()
            for j in range(i + 3, next_section_index, 2):
               node_ide = int(line_data[j])
               volume_fraction = float(line_data[j + 1])
               volume_fractions[node_ide] = volume_fraction

         except ValueError:
            raise ValueError('A FGM 3D Section does not have the necessary data in .dat file definition.')

         # Cadastrando Seção
         section = FGM_3D_Section(material_ide, volume_fractions)
         self.model.sections[section_ide] = section

         # Definindo Próximo Index
         i = next_section_index

   # Funções Privadas de Escrita de Seção
   def _write_section_him_3d(group: dict[int, HIM_3D_Section]) -> str:
      # Iniciando Output
      n_sections = len(group)
      span = len(str(max(group.keys())))
      output = f'\n%SECTION.HOMOGENEOUS.ISOTROPIC.3D\n{n_sections}\n'

      # Escrevendo Cada Seção
      for ide, section in group.items():
         output += f'{ide:<{span}}   {section.material_ide}\n'
      
      return output

   def _write_section_fgm_3d(group: dict[int, FGM_3D_Section]) -> str:
      # Iniciando Output
      n_sections = len(group)
      span = len(str(max(group.keys())))
      output = f'\n%SECTION.FGM.3D\n{n_sections}\n'

      # Escrevendo Cada Seção
      for ide, section in group.items():
         output += f'{ide:<{span}}   {section.material_ide}   {len(section.volume_fractions)}\n'
         
         # Escrevendo Cada Fração de Volume
         node_span = len(str(max(section.volume_fractions.keys())))
         for node_ide, volume_fraction in section.volume_fractions.items():
            output += f'{node_ide:<{node_span}}   {volume_fraction:.8e}\n'
      
      return output

   # Seções Suportadas
   supported_sections = {
      'HOMOGENEOUS.ISOTROPIC.3D': {
         'class': HIM_3D_Section,
         'read': _read_section_him_3d,
         'write': _write_section_him_3d
      },
      'FGM.3D': {
         'class': FGM_3D_Section,
         'read': _read_section_fgm_3d,
         'write': _write_section_fgm_3d
      }
   }

   def __init__(self):
      self.model = SimulationModel()
      self.reference = searcher.get_database('translation_reference')['dat']
   
   def read_nodes(self, dat_data: str):
      # Identificando Nodes
      keyword_format = '%NODE\n\d+\n\n%NODE.COORD\n\d+\n([^%]*)'
      node = '([+-]?\d+.\d+e?[+-]?\d*)'
      line_format = f'(\d+)\s+{node}\s+{node}\s+{node}'

      # Inserindo Nodes
      lines_data = re.findall(keyword_format, dat_data)[0]
      nodes = re.findall(line_format, lines_data)
      for node in nodes:
         ide, x, y, z = map(float, node)
         ide = int(ide)
         self.model.add_node(ide, x, y, z)
      
      # Identificando Pesos
      keyword_format = '%CONTROL.POINT.WEIGHT\n\d+\n([^%]*)'
      line_format = f'(\d+)\s+([+-]?\d+.\d+e?[+-]?\d*)'

      # Inserindo Nodes
      lines_data = re.findall(keyword_format, dat_data)
      if lines_data:
         lines_data = lines_data[0]
         weights = re.findall(line_format, lines_data)
         for node_ide, weight in weights:
            node_ide = int(node_ide)
            weight = float(weight)
            if weight == 1.0:
               continue
            self.model.nodes[node_ide].weight = weight
   
   def read_node_solver_order(self, dat_data: str) -> str:
      # Identificando Ordem de Resolução
      keyword_format = '%NODE.SOLVER.ORDER\n\d+\n([^%]*)'

      # Inserindo Ordem de Resolução
      node_ides = re.findall(keyword_format, dat_data)
      if len(node_ides) > 0:
         self.model.node_solver_order = [int(ide) for ide in node_ides[0].split()]

   def read_patches(self, dat_data: str):
      # Identificando Patches
      keyword_format = '%PATCH\n(\d+)\n([^%]*)'
      lines_data = re.findall(keyword_format, dat_data)

      # Verificando se Há Patches
      if len(lines_data) > 0:
         # Nomeando Dados
         n_patches = int(lines_data[0][0])
         lines_data = lines_data[0][1]
         supported_types = '|'.join(["'" + st + "'" for st in self.reference['patch_types']])
         patch_start_format = f'(\d+)\s+({supported_types})\s+1'

         # Separando Patches
         for _ in range(n_patches):
            # Localizando Dados Iniciais do Patch
            result = re.search(patch_start_format, lines_data)
            patch_ide, patch_type = result.groups()
            patch_ide = int(patch_ide)
            patch_type = patch_type[1:-1]
            index_start = result.end()

            # Determinando Geometria do Patch
            patch_geometry = self.reference['patch_types'][patch_type]

            # Localizando Dados Finais do Patch
            result = re.search(patch_start_format, lines_data[index_start:])
            index_end = None
            if result:
               index_end = result.start() + index_start
            
            # Lendo Knot Vectors
            patch_data = lines_data[index_start:index_end]
            knot_vectors = list()
            grade = list()
            while True:
               vector_format = "(\d+)\s+'General'\s+(\d+)"
               vector_result = re.search(vector_format, patch_data)

               # Parando Loop se Nenhum Knot Vector for Encontrado
               if not vector_result:
                  break

               # Atualizando Patch Data para Não incluir o Início do Patch Atual
               patch_data = patch_data[vector_result.end():]

               # Lendo Informações Básicas do Knot Vector
               grade_i, n_knots_i = vector_result.groups()
               grade_i = int(grade_i)
               n_knots_i = int(n_knots_i)

               # Fatiando Itens do Patch Data e Lendo Knots e Multiplicidades
               patch_data_splited = patch_data.split()
               knot_set = [float(patch_data_splited[i]) for i in range(n_knots_i)]
               knot_multiplicity = [int(patch_data_splited[i]) for i in range(n_knots_i, 2 * n_knots_i)]
               
               # Construindo Vetor de Knot
               knot_vector = []
               for knot, multiplicity in zip(knot_set, knot_multiplicity):
                  knot_vector += [knot] * multiplicity
               
               # Salvando Valores 
               knot_vectors.append(knot_vector)
               grade.append(grade_i)

            # Lendo Node Space
            node_space = [int(n) for n in patch_data_splited[2 * n_knots_i:]]

            # Calculando Número de Nodes por Elementos
            n_nodes = 1
            for p in grade:
               n_nodes *= p + 1

            # Adicionando Patch como Uma Geometria
            self.model.element_geometries[patch_ide] = ElementGeometry(
               shape = patch_geometry['shape'],
               base = patch_geometry['base'],
               grade = grade,
               n_nodes = n_nodes,
               n_dimensions = patch_geometry['n_dimensions'],
               knot_vectors =  knot_vectors,
               node_space = node_space
            )
            
            # Descartando Patch Localizado
            lines_data = lines_data[index_start:]

   def read_elements(self, dat_data: str):
      # Identificando Grupos de Elementos
      keyword_format = '%ELEMENT\.(.*)\n\d+\n([^%]*)'
      groups_data = re.findall(keyword_format, dat_data)

      # Analisando Cada Grupo
      group_ide = 1
      for element_type, lines_data in groups_data:
         # Dividindo Tipo e Teoria do Grupo de Elementos
         element_theory = None
         if element_type not in self.reference['elements']: 
            splited = element_type.split('.')
            if len(splited) > 1:
               # Tentando Identificar Teoria de Elemento
               element_theory = splited[0]
               try:
                  element_theory = self.reference['theories'][element_theory]
               except KeyError:
                  raise KeyError(f'The Element Theory "{element_theory}" is not supported for .dat files.')
               
               # Corrigindo Tipo de Elemento
               element_type = '.'.join(splited[1:])

         # Identificando Elementos
         try:
            type_info = self.reference['elements'][element_type]
         except KeyError:
            raise KeyError(f'The Element Type "{element_type}" is not supported for .dat files.')
         
         # Adaptando Leitura - Elementos de Bezier
         if type_info['base'] == 'Bezier':
            if type_info['shape'] == 'Triangle':
               group_ide = self._add_bezier_triangles(group_ide, lines_data, element_theory)
            elif type_info['shape'] == 'Quadrilateral':
               group_ide = self._add_bezier_surface(group_ide, lines_data, element_theory)
            else:
               raise KeyError(f'The Shape \"{type_info["shape"]}\" with Base \"{type_info["base"]}\" is not supported for .dat files.')
         
         # Adaptando Leitura - Elementos de BSpline
         elif type_info['base'] == 'BSpline':
            group_ide = self._add_bspline_elements(group_ide, lines_data, element_theory)
         
         # Adaptando Leitura - Elementos de Langrange
         else:
            int_ide = '(\d+)'
            node_ide = '\s+' + int_ide
            property_ides = '\s+\d+' * 2
            line_format = int_ide + property_ides + type_info['n_nodes'] * node_ide
            elements = re.findall(line_format, lines_data)

            # Criando Geometria
            geometry_ide = self.model.add_element_geometry(
               type_info['shape'],
               type_info['base'],
               type_info['grade'],
               type_info['n_nodes'],
               type_info['n_dimensions']
            )

            # Criando Grupo de Elementos
            self.model.add_element_group(group_ide, geometry_ide, element_theory)

            # Inserindo Elementos
            for element in elements:
               ide, *node_ides = map(int, element)
               self.model.add_element(group_ide, ide, node_ides)
         
         # Incrementando Ide do Grupo
         group_ide += 1
   
   def read_supports(self, dat_data: str):
      # Identificando Supports
      keyword_format = '%NODE\.SUPPORT\n\d+\n([^%]*)'
      bool_dof = '\s+(\d)'
      line_format = f'(\d+){bool_dof * 6}'

      # Inserindo Supports
      lines_data = re.findall(keyword_format, dat_data)
      if len(lines_data) > 0:
         supports = re.findall(line_format, lines_data[0])
         for support in supports:
            node_ide, *bool_dofs = support
            node_ide = int(node_ide)
            for index, bd in enumerate(bool_dofs):
               if bd == '1':
                  dof = self.model.supported_dofs[index]
                  self.model.add_support(node_ide, dof)

   def read_materials(self, dat_data: str):
      # Identificando Materiais
      keyword_format = '%MATERIAL\.(.+)\n\d+\n([^%]*)'

      # Inserindo Materiais
      lines_data = re.findall(keyword_format, dat_data)
      for line_data in lines_data:
         # Extraindo Informações
         type, line_data = line_data

         # Tentando Ler Material de Acordo com o Tipo
         try:
            read_function = DAT_Interpreter.supported_materials[type]['read']
            read_function(self, line_data)
         except KeyError:
            raise KeyError(f'The Material Type "{type}" is not supported for .dat files.')

   def read_sections(self, dat_data: str):
      # Identificando Seções
      keyword_format = '%SECTION\.(.+)\n\d+\n([^%]*)'

      # Inserindo Seções
      lines_data = re.findall(keyword_format, dat_data)
      for line_data in lines_data:
         # Extraindo Informações
         type, line_data = line_data

         # Tentando Ler Seção de Acordo com o Tipo
         try:
            read_function = DAT_Interpreter.supported_sections[type]['read']
            read_function(self, line_data)
         except KeyError:
            raise KeyError(f'The Section Type "{type}" is not supported for .dat files.')

   def read(self, dat_data: str):
      # Interpretando Nodes
      self.read_nodes(dat_data)

      # Interpretando Ordem de Resolução
      self.read_node_solver_order(dat_data)

      # Interpretando Patches
      self.read_patches(dat_data)

      # Interpretando Elementos
      self.read_elements(dat_data)

      # Interpretando Supports
      self.read_supports(dat_data)

      # Interpretando Materiais
      self.read_materials(dat_data)

      # Interpretando Seções
      self.read_sections(dat_data)
   
   def write_nodes(self) -> str:
      # Parâmetros Iniciais
      n_nodes = len(self.model.nodes)
      span = len(str(n_nodes))
      output = f'\n%NODE\n{n_nodes}\n\n%NODE.COORD\n{n_nodes}\n'

      # Escrevendo Cada Node
      weighted_nodes = list()
      for ide, node in self.model.nodes.items():
         offset = span - len(str(ide))
         offset = ' ' * offset
         output += f'{ide}{offset}   {node.x:+.8e}   {node.y:+.8e}   {node.z:+.8e}\n'

         # Verificando se Node tem Peso
         if node.weight:
            weighted_nodes.append(ide)

      # Escrevendo Pesos
      n_weights = len(weighted_nodes)
      if n_weights > 0:
         max_width = len(str(n_weights))
         output += f'\n%CONTROL.POINT.WEIGHT\n{n_weights}\n'
         for ide in weighted_nodes:
            output += f'{ide:<{max_width}}   {self.model.nodes[ide].weight:.6e}\n'
      
      return output
   
   def write_node_solver_order(self) -> str:
      # Parâmetros Iniciais
      solver_order = self.model.node_solver_order
      n = len(solver_order)
      output = f'\n%NODE.SOLVER.ORDER\n{n}\n'

      # Escrevendo Ordem
      max_width = len(str(n))
      for index in range(0, len(solver_order), 15):
         output += ' '.join([f'{node_ide:>{max_width}}' for node_ide in solver_order[index:index + 15]])
         output += '\n'
      
      return output

   def write_patches(self) -> str:
      # Procurando uma Geometria de base BSpline
      patch_ides = list()
      for ide, geometry in self.model.element_geometries.items():
         if geometry.base == 'BSpline':
            patch_ides.append(ide)
      
      # Escrevendo Patches se Houver Necessidade
      output = ''
      n_patches = len(patch_ides)
      if n_patches > 0:
         # Escrevendo Keyword do Patch
         output += f'\n%PATCH\n{n_patches}\n'
         
         # Escrevendo cada Patch
         for patch_ide in patch_ides:
            # Determinando Tipo do Patch
            patch_type = ''
            patch_geometry = self.model.element_geometries[patch_ide]
            for type, patch_info in self.reference['patch_types'].items():
               if (
                  patch_info['shape'] == patch_geometry.shape and
                  patch_info['n_dimensions'] == patch_geometry.n_dimensions
               ):
                  patch_type = type
                  break
            else:
               raise ValueError(f'Patch Type for "BSpline {patch_geometry.shape}" with {patch_geometry.n_dimensions} dimension(s) is not supported for writing.')
            
            # Escrevendo Definição Básica de um Patch
            output += f'{patch_ide} \'{patch_type}\' 1\n'

            # Escrevendo Vetores de Knot
            for degree, knot_vector in zip(patch_geometry.grade, patch_geometry.knot_vectors):
               knot_set = list(set(knot_vector))
               knot_set.sort()
               n_knot_set = len(knot_set)
               knot_multiplicity = ' '.join([str(knot_vector.count(k)) for k in knot_set])
               knot_set = '\n'.join([f'{k:.6e}' for k in knot_set])
               output += f'{degree}   \'General\'   {n_knot_set}\n{knot_set}\n{knot_multiplicity}\n'
            
            # Escrevendo Nodes
            max_width = len(str(len(self.model.nodes)))
            for index in range(0, len(patch_geometry.node_space), 15):
               output += ' '.join([f'{node_ide:>{max_width}}' for node_ide in patch_geometry.node_space[index:index + 15]])
               output += '\n'

      return output


   def write_elements(self) -> str:
      # Parâmetros Iniciais
      output = ''
      total_elements = 0
      n_nodes = len(self.model.nodes)
      node_ide_span = len(str(n_nodes))

      # Escrevendo Cada Grupo de Elemento
      for group in self.model.element_groups.values():
         # Parâmetros Iniciais
         n_elements = len(group.elements)
         total_elements += n_elements
         span = len(str(n_elements))

         # Buscando Tipo de Elemento Correspondente às Propriedades do Elemento
         element_type = ''
         geometry = self.model.element_geometries[group.geometry_ide]

         # Pesquisando Label - Elementos de Lagrange
         if geometry.base == 'Lagrange':
            for reference_type, reference_geometry in self.reference['elements'].items():
               if (
                  reference_geometry['shape'] == geometry.shape and
                  reference_geometry['base'] == geometry.base and
                  reference_geometry['grade'] == geometry.grade and
                  reference_geometry['n_nodes'] == geometry.n_nodes and
                  reference_geometry['n_dimensions'] == geometry.n_dimensions
               ):
                  element_type = reference_type
                  break
            else:
               raise ValueError(f'The "{geometry.base} {geometry.shape}" Geometry with grade {geometry.grade} and {geometry.n_nodes} nodes and {geometry.n_dimensions} dimensions is not supported for .dat files.')
         
         # Pesquisando Label - Elementos de Bezier e BSpline
         else:
            for reference_type, reference_geometry in self.reference['elements'].items():
               if (
                  reference_geometry['shape'] == geometry.shape and
                  reference_geometry['base'] == geometry.base and
                  reference_geometry['n_dimensions'] == geometry.n_dimensions
               ):
                  element_type = reference_type
                  break
            else:
               raise ValueError(f'The "{geometry.base} {geometry.shape}" Geometry with {geometry.n_dimensions} dimensions is not supported for .dat files.')

         # Verificando se Elemento Tem uma Teoria
         if group.theory:
            for dat_theory, reference_theory in self.reference['theories'].items():
               if reference_theory == group.theory:
                  element_type = f'{dat_theory}.{element_type}'
                  break
            else:
               raise ValueError(f'The Theory "{group.theory}" is not supported for .dat files.')

         output += f'\n%ELEMENT.{element_type}\n{n_elements}\n'

         # Escrevendo Cada Elemento - Elementos de Lagrange
         if geometry.base == 'Lagrange':
            # Alterando Informações para casos Especiais
            if geometry.shape == 'Line':
               more_info = '1'
            else:
               more_info = '1  1'
            
            for ide, element in group.elements.items():
               offset = span - len(str(ide))
               offset = ' ' * offset
               node_ides = '   '.join([ f'{nis:>{node_ide_span}}' for nis in element.node_ides ])
               output += f'{ide}{offset}   {more_info}   {node_ides}\n'
         
         # Escrevendo Cada Elemento - Elementos de Bezier
         elif geometry.base == 'Bezier':
            more_info = f'1  1  1  {geometry.grade}'
            for ide, element in group.elements.items():
               offset = span - len(str(ide))
               offset = ' ' * offset
               node_ides = '   '.join([ f'{nis:>{node_ide_span}}' for nis in element.node_ides ])
               output += f'{ide}{offset}   {more_info}   {node_ides}\n'
         
         # Escrevendo Cada Elemento - Elementos de Bezier
         elif geometry.base == 'BSpline':
            more_info = f'1  1  {group.geometry_ide}'
            for ide, element in group.elements.items():
               offset = span - len(str(ide))
               offset = ' ' * offset
               knot_spans = '   '.join([ f'{k:<{node_ide_span}}' for k in element.knot_span ])
               output += f'{ide}{offset}   {more_info}   {knot_spans}\n'

      output = f'\n%ELEMENT\n{total_elements}\n' + output
      return output
   
   def write_supports(self) -> str:
      # Parâmetros Iniciais
      n_supports = len(self.model.supports)
      span = len(str(n_supports))
      output = f'\n%NODE.SUPPORT\n{n_supports}\n'

      # Escrevendo Cada Support
      for node_ide, dofs in self.model.supports.items():
         offset = span - len(str(node_ide))
         offset = ' ' * offset
         dofs_str = [
            '1' if dof in dofs else '0' 
            for dof in self.model.supported_dofs
         ]
         dofs_str = ' '.join(dofs_str)
         output += f'{node_ide}{offset}   {dofs_str}\n'
      return output
   
   def write_materials(self) -> str:
      # Parâmetros Iniciais
      n_materials = len(self.model.materials)
      output = f'\n%MATERIAL\n{n_materials}\n'

      # Criando Grupos de Materiais
      groups = dict()
      for type in DAT_Interpreter.supported_materials.keys():
         groups[type] = dict()

      # Tentando Organizar Materiais em Grupos
      for ide, material in self.model.materials.items():
         # Verificando se o Material é Supportado
         for type, s in DAT_Interpreter.supported_materials.items():
            if s.get('class') and isinstance(material, s['class']):
               groups[type][ide] = material
               break
         else:
            raise TypeError(f'The Material Type "{material.__class__.__name__}" is not supported for .dat files.')
      
      # Escrevendo Grupos
      for type, group in groups.items():
         if len(group) > 0:
            output += self.supported_materials[type]['write'](group)

      return output
   
   def write_sections(self) -> str:
      # Parâmetros Iniciais
      n_sections = len(self.model.sections)
      output = f'\n%SECTION\n{n_sections}\n'

      # Criando Grupos de Seções
      groups = dict()
      for type in DAT_Interpreter.supported_sections.keys():
         groups[type] = dict()

      # Tentando Organizar Seções em Grupos
      for ide, section in self.model.sections.items():
         # Verificando se a Seção é Supportada
         for type, s in DAT_Interpreter.supported_sections.items():
            if s.get('class') and isinstance(section, s['class']):
               groups[type][ide] = section
               break
         else:
            raise TypeError(f'The Section Type "{section.__class__.__name__}" is not supported for .dat files.')
      
      # Escrevendo Grupos
      for type, group in groups.items():
         if len(group) > 0:
            output += self.supported_sections[type]['write'](group)

      return output

   def write(self) -> str:
      # Inicializando Output
      output = '%HEADER\n'

      # Escrevendo Nodes
      if len(self.model.nodes) > 0:
         output += self.write_nodes()

      # Escrevendo Supports
      if len(self.model.supports) > 0:
         output += self.write_supports()
      
      # Escrevendo Materiais
      if len(self.model.materials) > 0:
         output += self.write_materials()

      # Escrevendo Seções
      if len(self.model.sections) > 0:
         output += self.write_sections()

      # Escrevendo Ordem de Resolução
      if len(self.model.node_solver_order) > 0:
         output += self.write_node_solver_order()

      # Escrevendo Patches
      output += self.write_patches()

      # Escrevendo Elementos
      if len(self.model.element_groups) > 0:
         output += self.write_elements()

      # Finalizando Output
      output += '\n%END'
      
      return output

class SVG_Interpreter:
   def __init__(self):
      self.model = SimulationModel()
      self.node_radius = 1
      self.node_color = '#a95e5e'
      self.element_color = '#fcff5e'
      self.element_stroke_width = 1
      self.element_stroke_color = 'black'
      self.gt = GeometricalTransformer()
   
   def tesselate_bezier_curve(self, grade: int, points: list[Node], n_regions: int):
      # Variáveis Iniciais
      tesselated_points = list()
      p = grade
      h = 1 / (n_regions - 1)

      # Gerando Pontos da Curva
      for nr in range(n_regions):
         # Calculando Região do Espaço Paramétrico
         t = nr * h
         
         # Calculando Ponto Cartesiano Correspondente
         weight_sum, coord_x, coord_y = 0, 0, 0
         for point, i in zip(points, range(0, p + 1)):
            bp = bernstein_polynomial(i, p, t)
            w = point.weight or 1
            weight_sum += bp * w
            coord_x += bp * point.x * w
            coord_y += bp * point.y * w
         coord_x /= weight_sum
         coord_y /= weight_sum
         tesselated_points.append([coord_x, coord_y])
      
      # Corrigindo Pontos Ímpares para Coordenada Equivalente na Representação de Curva de Bezier Quadrática
      for i in range(1, len(tesselated_points), 2):
         tesselated_points[i][0] = bezier_equiv_coord(tesselated_points[i][0], tesselated_points[i - 1][0], tesselated_points[i + 1][0])
         tesselated_points[i][1] = bezier_equiv_coord(tesselated_points[i][1], tesselated_points[i - 1][1], tesselated_points[i + 1][1])
      
      # Retornando Pontos Tesselados (Excluindo o Primeiro)
      return tesselated_points[1:]

   def write_nodes(self) -> str:
      # Inicializando Node Output
      output = f'\n   <g id="Nodes" fill="{self.node_color}">'

      # Escrevendo Cada Node
      for node in self.model.nodes.values():
         output += f'\n      <circle cx="{node.x:.8e}" cy="{node.y:.8e}" r="{self.node_radius}" />'
      
      output += '\n   </g>'
      return output
   
   def write_bezier_triangles(self, grade: int, group: ElementGroup) -> str:
      # Parâmetros Iniciais
      output = ''
      p = grade
      nodes_total = int(3 + 3 * (p - 1) + ((p - 2) * (p - 1) / 2))
      indexes_corner = [1, nodes_total - p, nodes_total]

      # Index dos Nodes Intermediários
      ie1 = [int(1 + ((i + 1) * (i + 2) / 2)) for i in range(p - 1)]
      ie2 = [nodes_total - p + 1 + i for i in range(p - 1)]
      ie3 = [int((i + 2) * (i + 3) / 2) for i in range(p - 1)]
      ie3.reverse()
      indexes_by_edge = [ie1, ie2, ie3]

      # Escrevendo Path de Cada Elemento
      for element in group.elements.values():
         # Inicializando Path
         output += f'\n      <path d="'

         # Lado 1 - Ponto Incial
         node_corner_1 = self.model.nodes[element.node_ides[indexes_corner[0] - 1]]
         output += f'M {node_corner_1.x:.8e} {node_corner_1.y:.8e} '

         # Construindo Curvas de Bezier para Cada Lado
         for indexes_edge, index_corner in zip(indexes_by_edge, indexes_corner[1:] + [indexes_corner[0]]):
            # Obtendo Pontos do Lado
            node_corner_2 = self.model.nodes[element.node_ides[index_corner - 1]]
            points = [self.model.nodes[element.node_ides[i - 1]] for i in indexes_edge]
            points.append(node_corner_2)
            points.insert(0, node_corner_1)

            # Calculando Fator de Colinearidade dos Pontos
            c_factor = self.gt.calculate_colinearity([[p.x, p.y] for p in points])

            # Resumindo Path em Uma linha reta para um fator baixo
            if c_factor < 0.1:
               output += f'L {node_corner_2.x:.8e} {node_corner_2.y:.8e} '

            # Tesselando Curva com Base no Fator
            else:
               # Definindo Discretização da Tesselação com Base no Fator de Colinearidade
               n_regions = (2 * p - 1) + (2 * floor(c_factor / 50))

               # Gerando Pontos de Tesselação
               tp = self.tesselate_bezier_curve(p, points, n_regions)

               for i in range(0, len(tp), 2):
                  output += f'Q {tp[i][0]:.8e} {tp[i][1]:.8e}, {tp[i + 1][0]:.8e} {tp[i + 1][1]:.8e} '
            
            node_corner_1 = node_corner_2

         output += 'Z" />'
      
      return output
   
   def write_bspline_surface(self, geometry: ElementGeometry, group: ElementGroup) -> str:
      # Montando Matriz de Identificadores do Pontos de Controle
      output = ''
      control_points_ides = list()
      n_basis_u = len(geometry.knot_vectors[0]) - geometry.grade[0] - 1
      n_basis_v = len(geometry.knot_vectors[1]) - geometry.grade[1] - 1
      for i in range(n_basis_u):
         control_points_ides.append(geometry.node_space[i::n_basis_u])
      
      # Criando Pontos de Controle e Pesos
      control_points = [
         [
            [
               self.model.nodes[control_points_ides[i][j]].x,
               self.model.nodes[control_points_ides[i][j]].y
            ]
            for j in range(n_basis_v)
         ]
         for i in range(n_basis_u)
      ]
      weights = [
         [
            self.model.nodes[control_points_ides[i][j]].weight or 1.0
            for j in range(n_basis_v)
         ]
         for i in range(n_basis_u)
      ]

      # Criando Superfície NURBS
      nurbs_surface = NURBS_Surface(
         degree = geometry.grade,
         knot_vectors = geometry.knot_vectors,
         control_points = control_points,
         weights = weights
      )

      # Criando Intervalos Paramétricos de Elementos
      d_u, d_v = geometry.grade
      d_u *= 2
      d_v *= 2
      k_u = sorted(list(set(geometry.knot_vectors[0])))
      k_v = sorted(list(set(geometry.knot_vectors[1])))
      for j in range(len(k_v) - 1):
         v1, v2 = k_v[j:j + 2]
         v = [v1 + k * (v2 - v1) / (d_v) for k in range(d_v)]
         v.append(v2)
         for i in range(len(k_u) - 1):
            u1, u2 = k_u[i:i + 2]
            u = [u1 + k * (u2 - u1) / (d_u) for k in range(d_u)]
            u.append(u2)

            # Criando Pontos da Fronteira do Elemento
            edges = [
               [
                  nurbs_surface(ui, v1)
                  for ui in u
               ],
               [
                  nurbs_surface(u2, vi)
                  for vi in v
               ],
               [
                  nurbs_surface(ui, v2)
                  for ui in list(reversed(u))
               ],
               [
                  nurbs_surface(u1, vi)
                  for vi in list(reversed(v))
               ]
            ]

            # Iniciando Path do Elemento
            output += f'\n      <path d="'

            # Escrevendo Ponto Inicial
            output += f'M {edges[0][0][0]:.8e} {edges[0][0][1]:.8e} '

            # Percorendo Lados
            for edge in edges:
               # Resumindo Path em Uma linha reta para um fator baixo
               c_factor = self.gt.calculate_colinearity(edge)
               if c_factor < 0.1:
                  output += f'L {edge[-1][0]:.8e} {edge[-1][1]:.8e} '

               # Tesselando Curva se o Fator de Colinearidade for Alto
               else:
                  for index in range(1, len(edge) - 1, 2):
                     x_c = bezier_equiv_coord(edge[index][0], edge[index - 1][0], edge[index + 1][0])
                     y_c = bezier_equiv_coord(edge[index][1], edge[index - 1][1], edge[index + 1][1])
                     output += f'Q {x_c:.8e} {y_c:.8e}, {edge[index + 1][0]:.8e} {edge[index + 1][1]:.8e} '
            
            # Finalizando Path
            output += 'Z" />'
      return output

   def write_finite_elements(self, grade: int, group: ElementGroup) -> str:
      output = ''

      # Tratamento para Elementos Lineares
      if grade == 1:
         for element in group.elements.values():
            output += '\n      <polygon points="'

            # Escrevendo Cada Ponto
            for ide in element.node_ides:
               node = self.model.nodes[ide]
               output += f'{node.x:.8e},{node.y:.8e} ' 
            output += '" />'

      # Tratamento para Elementos Quadráticos
      else:
         for element in group.elements.values():
            # Escrevendo Ponto Inicial
            node = self.model.nodes[element.node_ides[0]]
            output += f'\n      <path d="M {node.x:.8e} {node.y:.8e} '

            # Escrevendo Lados como Curvas Quadráticas de Bezier
            for i in list(range(2, len(element.node_ides), 2)) + [0]:
               n2 = self.model.nodes[element.node_ides[i]]
               nc = self.model.nodes[element.node_ides[i - 1]]
               n0 = self.model.nodes[element.node_ides[i - 2]]
               x1 = bezier_equiv_coord(nc.x, n0.x, n2.x)
               y1 = bezier_equiv_coord(nc.y, n0.y, n2.y)
               output += f'Q {x1:.8e} {y1:.8e}, {n2.x:.8e} {n2.y:.8e} ' 
            output += 'Z" />'
      return output

   def write_elements(self) -> str:
      # Inicializando Node Output
      output = f'\n   <g id="Elements" fill="{self.element_color}" stroke="{self.element_stroke_color}" stroke-width="{self.element_stroke_width}">'

      # Escrevendo Cada Grupo de Elemento
      for group in self.model.element_groups.values():
         # Identificando Geometria do Grupo
         geometry = self.model.element_geometries[group.geometry_ide]

         # Tratamento para Elementos de Bezier
         if geometry.shape == 'Triangle' and geometry.base == 'Bezier':
            output += self.write_bezier_triangles(geometry.grade, group)
         
         # Tratamento para Superfícies B-Spline
         elif geometry.shape == 'Quadrilateral' and geometry.base == 'BSpline':
            output += self.write_bspline_surface(geometry, group)

         # Tratamento para Elementos Finitos Tradicionais
         else:
            output += self.write_finite_elements(geometry.grade, group)

      output += '\n   </g>'
      return output

   def write(self) -> str:
      # Inicializando Output
      output = '<svg width="100" height="100" version="1.1" xmlns="http://www.w3.org/2000/svg">'

      # Calculando Raio dos Nodes e Largura do Delinado dos Elementos Ideais
      self.node_radius = 9.5 / (len(self.model.nodes) - 1) ** 0.5 + 0.1
      self.element_stroke_width = self.node_radius * 0.5

      # Escrevendo Elementos
      output += self.write_elements()

      # Escrevendo Nodes
      output += self.write_nodes()

      # Finalizando Output
      output += '\n</svg>'
      
      return output


from ..interface import searcher, messenger
from ..models.custom_errors import CommandError
from ..models.ansi_style import Color, Effect

# Pesquisando Mensagens de Ajuda na Base de Dados
help_messages = searcher.get_database('help_messages')

# Função de Inicialização
def start(command_name: str = 'default'):   
   # Coletando Dados da Mensagem de Ajuda com base nos Argumentos.
   try:
      message = help_messages[command_name]
   except KeyError:
      raise CommandError(f'There is no command named "{command_name}".')
   
   # Gerando Mensagem
   messenger.show(f"{Effect.bold}Usage:{Effect.reset} {message['usage']}\n")
   for title, points in message['topics'].items():
      # Exibindo Título do Tópico
      messenger.show(f"{Color.green + Effect.bold}{title}:{Effect.reset}")

      # Determinando Largura Máxima da Primera Coluna dos Pontos do Tópico
      max_width = 0
      for name in points.keys():
         len_name = len(name)
         if len_name > max_width:
            max_width = len_name
      max_width += 3

      # Exibindo Pontos do Tópico
      for name, description in points.items():
         # Dividindo Descrição em Linhas
         lines = description.split('. ')

         # Cortando Primeira Linha se Necessário
         max_size = 79 - (max_width + 5)
         first_line = lines.pop(0)
         
         if (first_line_cropped := len(first_line) > max_size):
            crop_index = first_line[:max_size].rfind(' ')
            lines.insert(0, first_line[crop_index + 1:])
            first_line = first_line[:crop_index]
         
         # Exibindo a Primeira Linha
         first_line = f"\n{name:<{max_width}}|   {first_line}"
         if not first_line_cropped:
            first_line += '.'
         messenger.show(first_line)

         # Exibindo Demais Linhas
         for line in lines:
            # Quebrando Linhas para Terem 80 Caracteres
            offset = ' ' * max_width
            while len(line) > max_size:
               crop_index = line[:max_size].rfind(' ')
               cropped_line = line[:crop_index]
               line = line[crop_index + 1:]
               messenger.show(f"{offset}|   {cropped_line}")
            messenger.show(f"{offset}|   {line}.")

      # Quebra de Linha Final do Tópico
      messenger.show('')


# Funções Globais
def read(file_path: str) -> str:
   try:
      file = open(file_path, 'r')
      file_data = file.read()
   except OSError:
      raise OSError(f'Could not open or read the {file_path}')
   file.close()
   return file_data

def write(file_path: str, file_data: str):
   try:
      file = open(file_path, 'w')
      file.write(file_data)
   except OSError:
      raise OSError(f'Could not write the {file_path}')
   file.close()